
from FITS import *
from numpy.oldnumeric.mlab import diff
from VTKHelperFunctions import *
from basics import *
from numpy.core.numerictypes import *
import numpy
from sets import *

import ZODB, ZODB.FileStorage, transaction
import persistent
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree
from persistent.dict import *

from GalPhot import *

import ImageDraw
import Image

from SimpleQueue import *

def WriteDictionary(zdb,dict):
    try:
       if os.path.exists(zdb): os.remove(zdb)
       if os.path.exists(zdb+".tmp"): os.remove(zdb+".tmp")
       if os.path.exists(zdb+".index"): os.remove(zdb+".index")
       if os.path.exists(zdb+".lock"): os.remove(zdb+".lock")
    except:
       print "Odd, we couldn't remove old zdb stuff?"

    storage = ZODB.FileStorage.FileStorage(zdb)
    print "Writing to",zdb
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    for key in dict:
        dbroot[key] = dict[key]
    transaction.commit()
    conn.cacheMinimize()

    conn.close()
    db.close()
    storage.close()

def ReadDictionary(zdb,verb=0,noclose=False):
    t0 = time.time()
    storage = ZODB.FileStorage.FileStorage(zdb,read_only=True)
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    ck = list(dbroot.keys())
    dict = {}
    for j in ck:
        dict[j] = dbroot[j]
        if verb:
           print j,
           sys.stdout.flush()
    if noclose:
       return dict,conn,db,storage
    else:
       conn.close()
       db.close()
       storage.close()
       t1 = time.time()
       print "Closed! Time to read %s is %.3s."%(zdb,t1-t0)
       sys.stdout.flush()
       return dict


def WriteContours(zdb,contours):
    try:
       if os.path.exists(zdb): os.remove(zdb)
       if os.path.exists(zdb+".tmp"): os.remove(zdb+".tmp")
       if os.path.exists(zdb+".index"): os.remove(zdb+".index")
       if os.path.exists(zdb+".lock"): os.remove(zdb+".lock")
    except:
       print "Odd, we couldn't remove old zdb stuff?"

    storage = ZODB.FileStorage.FileStorage(zdb)
    print "Writing to",zdb
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    dbroot["Contours"] = OOBTree()
    if isinstance(contours,ListType):
       items = range(len(contours))
    else:
       items = contours.keys()
       items.sort()
    for j in items:
        dbroot["Contours"][j] = copy.deepcopy(contours[j])
    transaction.commit()
    conn.cacheMinimize()

    conn.close()
    db.close()
    storage.close()

def ReadContours(zdb,verb=0,noclose=False):
    storage = ZODB.FileStorage.FileStorage(zdb,read_only=True)
    db = ZODB.DB(storage)
    conn = db.open()
    dbroot = conn.root()
    contours = {}
    ck = list(dbroot["Contours"].keys())
    print "Reading %s(%d)..." % (zdb,len(ck)),
    sys.stdout.flush()
    for j in ck:
        if noclose:
           contours[j] = dbroot["Contours"][j]
        else:
           contours[j] = copy.deepcopy(dbroot["Contours"][j])
        if verb:
           print j,
           sys.stdout.flush()
    if noclose:
       return contours,conn,db,storage
    else:
       conn.close()
       db.close()
       storage.close()
       print "Closed!"
       sys.stdout.flush()
       return contours

def MakeContourHash(contours,dx=10,dy=10):
    F,B = {},{}
    print "Making hash table!"
    t0 = time.time()
    for contour in contours:
        xi = floor(contour.xb/dx+0.5).astype(int32)
        yi = floor(contour.yb/dy+0.5).astype(int32)
        xyi = Set([(x1,y1) for x1,y1 in zip(xi,yi)])
        [F.__setitem__(xy,F.get(xy,[])+[contour]) for xy in xyi]
        B[contour] = xyi
    for xy in F:
        F[xy] = Set(F[xy])
    t1 = time.time()
    print len(B),len(F)
    print "Done making hash table in %.3fs"%(t1-t0)
    return F,B

class Contour(persistent.Persistent):
    def __init__(self,contour,v):
        self.contour = contour
        self.v = v
        (self.xb,self.yb),(self.xe,self.ye) = transpose(contour)
        self.x = concatenate([self.xb,self.xb[:1]])
        self.y = concatenate([self.yb,self.yb[:1]])
#        self.getcenter()
        self.meanx,self.meany = mean(self.x),mean(self.y)
        self.medianx,self.mediany = median(self.x),median(self.y)
        self.mx,self.my = self.meanx,self.meany
        self.r = sqrt(sum((self.x-self.mx)**2 + (self.y-self.my)**2)/len(self.x))
        self.ex = extrema(self.xb)
        self.ey = extrema(self.yb)
        self.wid = diff(self.ex)[0]
        self.hei = diff(self.ey)[0]
        self.box = self.wid*self.hei
        x1,y1 = transpose(asarray(contour)[::,::,0])
        x2,y2 = transpose(asarray(contour)[::,::,1])
        self.area = 0.5*abs(sum(x1*y2-y1*x2))
        self.re = sqrt(self.area/pi)
        self.moments()
    def getcenter(self):
        segments = self.contour
        for j in range(1):
            (xb,yb),(xe,ye) = transpose(segments)
            x = 0.5*(xb+xe)
            y = 0.5*(yb+ye)
            mx,my = median(x),median(y)
#            print extrema(x),mx,extrema(y),my
            x = concatenate([x,x[:1]])
            y = concatenate([y,y[:1]])
            segments = asarray([[(x[j],x[j+1]),(y[j],y[j+1])] for j in range(len(x)-1)])
        self.mx,self.my = mx,my
    def multiply(self,fac,mx=None,my=None):
        (xb,yb),(xe,ye) = transpose(self.contour)
        x = concatenate([xb,xb[:1]])
        y = concatenate([yb,yb[:1]])
        if mx == None and my == None: mx,my = self.mx,self.my
        xb,xe = fac*(xb-mx)+mx, fac*(xe-mx)+mx
        yb,ye = fac*(yb-my)+my, fac*(ye-my)+my
        contour = transpose([[xb,yb],[xe,ye]])
        return Contour(contour,self.v)
    def distance(self,xa,ya):
        dx = xa[::,NewAxis]-self.x[NewAxis,::]
        dy = ya[::,NewAxis]-self.y[NewAxis,::]
        ds = sqrt(dx*dx+dy*dy)
        ds = array([min(ds1) for ds1 in ds])
        return ds
    def inside(self,xa,ya):
        t = Interiors(xa,ya,self.contour,verb=0)
        interior = less(abs(t-2*pi),1e-4)
#        ae = arctan2(self.ye[::,NewAxis]-ya[NewAxis,::],self.xe[::,NewAxis]-xa[NewAxis,::])
#        ab = arctan2(self.yb[::,NewAxis]-ya[NewAxis,::],self.xb[::,NewAxis]-xa[NewAxis,::])
#        t = add.reduce(mangle(ae-ab),0)
#        interior = less(abs(t-2*pi),1e-4)
        return interior
    def pixels(self,data):
        x0,x1 = int(self.ex[0]), int(self.ex[1]+1)
        y0,y1 = int(self.ey[0]), int(self.ey[1]+1)
        xpixels, ypixels, zpixels = [],[],[]
        for y in arange(y0,y1,1,int32):
            ua = between(self.y,y-2,y+2,inc=1)
            if ua.any():
               x0,x1 = extrema(compress(ua,self.x))
               x0,x1 = clip([int(x0-1),int(x1+2)],0,data.shape[1])
               xa = arange(x0,x1,1,float32)
               ya = y*ones(xa.shape)
               interior = self.inside(xa,ya)
               if interior.any():
                  za = data[y,x0:x1]
                  ya,xa,za = compress(interior,[ya,xa,za],1)
                  ypixels = concatenate([ypixels,ya])
                  xpixels = concatenate([xpixels,xa])
                  zpixels = concatenate([zpixels,za])
        return asarray([xpixels, ypixels, zpixels])
    def moments(self,x=None,y=None,w=None):
        if w == None: w = ones(self.x.shape,float)
        if x == None: xi = self.x
        else: xi = x
        if y == None: yi = self.y
        else: yi = y
        mx = divz(sum(xi*w),sum(w))
        my = divz(sum(yi*w),sum(w))
        x,y = xi-mx, yi-my
        mx2 = divz(sum(x**2*w),sum(w))
        my2 = divz(sum(y**2*w),sum(w))
        mxy = divz(sum(x*y*w),sum(w))
        th = 0.5*arctan2(2*mxy,mx2-my2)
        cth,sth = cos(th), sin(th)
        x,y = x*cth + y*sth, -x*sth + y*cth
        sx2 = sqrt(2*divz(sum(x**2*w),sum(w)))
        sy2 = sqrt(2*divz(sum(y**2*w),sum(w)))
        self.momx = mx
        self.momy = my
        self.moma = sx2
        self.momb = sy2
        self.cth = cos(th)
        self.sth = sin(th)
        self.momth = th*180/pi
        self.momab = getsqrt(self.momb*self.moma)
        self.momba = divz(self.momb,self.moma)
    def model(self,dth=3.0):
        th = arange(0.0,360.0,dth,float)
        if th[-1] != 360.0: th = concatenate([th,[360.0]])
        cth,sth = cos(th*pi/180), sin(th*pi/180)
        x,y = self.moma*cth, self.momb*sth 
        self.ellx,self.elly = x*self.cth - y*self.sth + self.momx, \
                x*self.sth + y*self.cth + self.momy
        xb,xe = self.ellx[:-1], self.ellx[1:]
        yb,ye = self.elly[:-1], self.elly[1:]
        contour = transpose([[xb,yb],[xe,ye]])
        return Contour(contour,self.v)


class Ellipse(persistent.Persistent):
    def __init__(self,x,y,w,verb=0,fit=1,model=1,dth=3.0,ns=0,shape=None):
        self.x = x
        self.y = y
        self.w = w
        self.mx = mean(x)
        self.my = mean(y)
        self.ns = ns
        self.v = mean(self.w)
        self.dth = dth
        self.shape = shape
        self.verb = verb
        if fit: self.fit()
        if model and self.mx!=99:
           self.contour = self.model()
        else: self.fail = True
    def __repr__(self):
        if hasattr(self,"ellipse"):
            fmt = " ".join(7*["%10.3f"])
            return fmt % tuple(self.ellipse)
        else: return "ellipse"
    def fit(self):
        self.mx =-99
        if len(self.x):
           wts = ones(self.x.shape)
           ex = diff(extrema(self.x))[0]
           ey = diff(extrema(self.y))[0]
           x,y,w = self.x, self.y, self.w
           ellipse = FitEllipse(x,y,w)
           self.ellipse = ellipse
           fmt = " ".join(7*["%10.3f"])
           if self.verb:
              print fmt % tuple(ellipse)
           self.mx,self.my,self.f,self.r,self.theta,self.a,self.b = self.ellipse
           xf = self.xmodel(self.mx,self.my,self.a,self.b,self.theta)
           yf = self.ymodel(self.mx,self.my,self.a,self.b,self.theta)
           self.rx, self.ry = rms(xf-self.x), rms(yf-self.y)
           self.ab = sqrt(self.b*self.a)
           self.ba = divz(self.b/self.a)
           self.area = pi*self.a*self.b
           self.cth = cos(self.theta*pi/180)
           self.sth = sin(self.theta*pi/180)
        if self.mx == -99 or self.a <= 0 or self.b <= 0 or \
                self.ab <= 0 or self.a > 2*max([ex,ey]):
           self.fail=True
           self.ab = -99
        else:
           self.fail = False
    def ymodel(self,mx,my,a,b,theta):
        cth,sth = cos(theta*pi/180), sin(theta*pi/180)
        x,y = self.x-mx,self.y-my
        x,y = x*cth + y*sth, -x*sth + y*cth
        y1 = b*getsqrt(1-power(x/a,2))
        y2 = -b*getsqrt(1-power(x/a,2))
        y = where(greater(y,0),y1,y2)
        x,y = x*cth - y*sth + mx, x*sth + y*cth + my
        return y
    def xmodel(self,mx,my,a,b,theta):
        cth,sth = cos(theta*pi/180), sin(theta*pi/180)
        x,y = self.x-mx,self.y-my
        x,y = x*cth + y*sth, -x*sth + y*cth
        x1 = a*getsqrt(1-power(y/b,2))
        x2 = -a*getsqrt(1-power(y/b,2))
        x = where(greater(x,0),x1,x2)
        x,y = x*cth - y*sth + mx, x*sth + y*cth + my
        return x
    def moments(self,w=None):
        if w == None: w = self.w
        else: w = w*self.w
        mx = divz(sum(self.x*w),sum(w))
        my = divz(sum(self.y*w),sum(w))
        x,y = self.x-mx, self.y-my
        mx2 = divz(sum(x**2*w),sum(w))
        my2 = divz(sum(y**2*w),sum(w))
        mxy = divz(sum(x*y*w),sum(w))
        th = 0.5*arctan2(2*mxy,mx2-my2)
        cth,sth = cos(th), sin(th)
        x,y = x*cth + y*sth, -x*sth + y*cth
        sx2 = sqrt(2*divz(sum(x**2*w),sum(w)))
        sy2 = sqrt(2*divz(sum(y**2*w),sum(w)))
        print mx,my,sx2,sy2,th*180/pi
    def model(self,mag=1):
        th = arange(0.0,360.0,self.dth,float)
        if th[-1] != 360.0: th = concatenate([th,[360.0]])
        cth,sth = cos(th*pi/180), sin(th*pi/180)
        x,y = mag*self.a*cth, mag*self.b*sth 
        self.ellx,self.elly = x*self.cth - y*self.sth + self.mx, \
                x*self.sth + y*self.cth + self.my
        xb,xe = self.ellx[:-1], self.ellx[1:]
        yb,ye = self.elly[:-1], self.elly[1:]
        if self.shape != None:
           xb,xe = clip([xb,xe],0,self.shape[1]-1)
           yb,ye = clip([yb,ye],0,self.shape[0]-1)
        contour = transpose([[xb,yb],[xe,ye]])
        return Contour(contour,self.v)

def RobustFitEllipse(x0,y0,w0,n):
    x0,y0,w0 = compress(w0,[x0,y0,w0],1)
    for i in range(2*n+1):
        j0 = i*l/2
        j1 = (i+1)*l/2
        x = x1[j0:j1]
        y = y1[j0:j1]
        w = w1[j0:j1]
        ellipse = FitEllipse(x,y,w)
        print j0,j1,len(x0),l,ellipse
        fits.append(ellipse)
    fits = asarray(fits)
    mfits = median(fits)
    print mfits
    print
    return mfits
    

def FitEllipse(Xt,Yt,Wt):
    f = sum(Wt)
    mx = sum(Wt*Xt)/f
    my = sum(Wt*Yt)/f
    sx = sqrt(2*sum(power(Xt-mx,2)*Wt)/f)
    sy = sqrt(2*sum(power(Yt-my,2)*Wt)/f)
    if not sx: sx = 1
    if not sy: sy = 1
    Xr = (Xt-mx)/sx
    Yr = (Yt-my)/sy
    D = transpose([Xr*Xr, Xr*Yr, Yr*Yr, Xr, Yr, ones(Xr.shape)])
#    mm = lambda a1,a2: matrixmultiply(a1,a2)
    mm = lambda a1,a2: dot(a1,a2)
    D = D*Wt[::,NewAxis]
    Dt = transpose(D)
    S = mm(Dt,D)
    C = zeros((6,6),float64)
    C[0,2] = -2.0
    C[1,1] = 1.0
    C[2,0] = -2.0
    tmpA = S[:3,:3]
    tmpB = S[:3,3:]
    tmpC = S[3:,3:]
    tmpD = C[:3,:3]
    if not determinant(tmpC): return -99,-99,-99,-99,-99,-99,-99
    tmpE = mm(inverse(tmpC),transpose(tmpB))
    eigval,eigvec = eigenvectors(mm(inverse(tmpD),(tmpA-mm(tmpB,tmpE))))
    neg = nonzero(less(eigval.real,0))
    if not len(neg): return -99,-99,-99,-99,-99,-99,-99
    ie = neg[0]
    A = eigvec[ie].real
    A = concatenate([A,mm(-tmpE,A)])
    par = asarray([ A[0]*sy*sy,
          A[1]*sx*sy,
          A[2]*sx*sx,
          -2*A[0]*sy*sy*mx - A[1]*sx*sy*my + A[3]*sx*sy*sy,
          -A[1]*sx*sy*mx - 2*A[2]*sx*sx*my + A[4]*sx*sx*sy,
          A[0]*sy*sy*mx*mx + A[1]*sx*sy*mx*my + A[2]*sx*sx*my*my
          - A[3]*sx*sy*sy*mx - A[4]*sx*sx*sy*my
          + A[5]*sx*sx*sy*sy ])
    theta = 0.5*arctan2(par[1],par[0]-par[2])
    cost = cos(theta); sint = sin(theta); sint2 = sint**2; cost2 = cost**2; sincost = sint*cost
    Ao = par[5]
    Au = par[3]*cost + par[4]*sint
    Av =-par[3]*sint + par[4]*cost
    Auu = par[0]*cost2 + par[2]*sint2 + par[1]*sincost
    Avv = par[0]*sint2 + par[2]*cost2 - par[1]*sincost
    if not Auu or not Avv: return -99,-99,-99,-99,-99,-99,-99
    tu = -Au/(2*Auu)
    tv = -Av/(2*Avv)
    wc = Ao - Auu*tu**2 - Avv*tv**2
    u = tu*cost - tv*sint
    v = tu*sint + tv*cost
    Ru = -wc/Auu
    Rv = -wc/Avv
    Ru = sqrt(abs(Ru))*sign(Ru)
    Rv = sqrt(abs(Rv))*sign(Rv)
    mx,my,f,r,theta,a,b = u,v,f,sqrt(Ru*Rv),theta*180/pi,Ru,Rv
    if b > a: a,b,theta = b,a,theta+90.0
    return mx,my,f,r,theta,a,b


def ContourFamily(contours):
    contours = {}
    ellipses = {}

    if all:
       for j2,n2 in enumerate(ns):
           contours[n2], ellipses[n2] = GetContours( d2(n2),n2,n0,levels[j2:])
    else:
       contours[n1], ellipses[n1] = GetContours( image2,n1,n0,levels)


    conall = {}
    ellall = {}
    for n2 in ns:
        if n2 in contours:
           ne = int(log(1.*n2/n0)/log(2))
           for level in levels:
               if level in contours[n2]:
                  for j in contours[n2][level]:
                      if not ellipses[n2][level][j].fail:
                         w = "%04d%05d%05d" % (n2,level,j)
                         conall[w] = contours[n2][level][j]
                         ellall[w] = ellipses[n2][level][j]

    ws = ellall.keys()
    print len(ws)
    ws = [w for w in ws if not ellall[w].fail]
    print len(ws)

    ab = array([ellall[w].ab for w in ws])
    area = array([conall[w].area for w in ws])
    print "Quartiles of sqrt(ab)=",quartiles(ab)
    print "Extrema of sqrt(ab)=",extrema(ab)
    print "Quartiles of area=",quartiles(area)
    print "Extrema of area=",extrema(area)
    order = argsort(area)[::-1]
    ws = [ws[j] for j in order]
#    ws = [w for w in ws if 2400<conall[w].mx<2600 and 2400<conall[w].my<2600]
#    ws = [w for w in ws if 2200<conall[w].mx<2800 and 2200<conall[w].my<2800]
#    ws = [w for w in ws if 2000<conall[w].mx<3000 and 2000<conall[w].my<3000]
#    ws = [w for w in ws if 1000<conall[w].mx<2000 and 2000<conall[w].my<3000]

    if 1:
       print "Get enclosed counts for",len(ws)
       x0s,x1s,y0s,y1s,boxes,areas = [],[],[],[],[],[]
       for j,w in enumerate(ws):
           x0,x1 = int(conall[w].ex[0]), int(conall[w].ex[1]+1)
           y0,y1 = int(conall[w].ey[0]), int(conall[w].ey[1]+1)
           box = (x1-x0)*(y1-y0)
           areas.append(conall[w].area)
           x0s.append(x0)
           x1s.append(x1)
           y0s.append(y0)
           y1s.append(y1)
           boxes.append(box)
       x0s,x1s,y0s,y1s,boxes,areas = asarray( [x0s,x1s,y0s,y1s,boxes,areas])
       print extrema(x0s), extrema(x1s), extrema(x1s-x0s)
       print extrema(y0s), extrema(y1s), extrema(y1s-y0s)
       print extrema(boxes),quartiles(boxes)
       print extrema(areas),quartiles(areas)

       t0 = time.time()
       for j,w in enumerate(ws):
           if j and not (j%(len(ws)/160+1)):
              sys.stdout.write(".")
              sys.stdout.flush()
           x0,x1 = int(conall[w].ex[0]), int(conall[w].ex[1]+1)
           y0,y1 = int(conall[w].ey[0]), int(conall[w].ey[1]+1)
           box = (y1-y0)*(x1-x0)
#           print "%s %9.1f %4d %4d %4d %4d %6d %7.4f" % (w,conall[w].area,x0,x1,y0,y1,box,1.0*conall[w].area/box,)
#           sys.stdout.flush()
           xpixels, ypixels, zpixels = conall[w].pixels(data)
           wpixels = divz(zpixels*greater(zpixels,0),sqrt((getsqrt(zpixels)**2+sigma**2)))
           upixels = greater(wpixels,0)
           xpixels,ypixels,wpixels,zpixels = compress(upixels, \
                   [xpixels,ypixels,wpixels,zpixels], 1)
           conall[w].counts = sum(zpixels)
           conall[w].ix = xpixels
           conall[w].iy = ypixels
           conall[w].iz = zpixels
           conall[w].flux = divz(conall[w].counts, conall[w].area)

#           if len(wpixels) > max([0.25*conall[w].area,pi]):
#             e2 = Ellipse(xpixels,ypixels,wpixels,ns=0,verb=0)
##             if diff(extrema(e2.ellx))[0]>0.75*(naxis1-2) or \
##                      diff(extrema(e2.elly))[0]>0.75*(naxis2-2):
##                e2.fail = True
#             if e2.fail:
#                conall[w].counts = 0.0
##             else:
##                print w, len(xpixels), conall[w].area, conall[w].ex, extrema(e2.ellx), conall[w].ey, extrema(e2.elly)
##                conall[w].counts = sum(zpixels)
##                val = mean(zpixels)-3*sigma
##                mag = divz(val*conall[w].area,conall[w].counts)
##           print conall[w].counts
##           sys.stdout.flush()
##           raw_input()

       t1 = time.time()
       print
       print "Time to get enclosed counts =",t1-t0

    print len(ws)
    ws = [w for w in ws if not ellall[w].fail and conall[w].counts>0]
    print len(ws)
    
    laps = {}
    print "Make families!",len(ws)
    used = {}
    for j,w in enumerate(ws[:-1]):
        if j and not (j%(len(ws)/80+1)): sys.stdout.write("."); sys.stdout.flush()
        lap1 = GetOverlaps(w,ws[j+1:],conall,ellall)
        used.setdefault(w,w)
        if lap1:
           u = [w2 for w2 in lap1 if w2 in used]
           if 0 and len(u):
              wa = list(Set([used[w2] for w2 in u]))
              w1 = wa[argsort([conall[w2].area for w2 in wa])[-1]]
              [laps[w1].setdefault(w2) for w2 in lap1]
              [used.__setitem__(w2,w1) for w2 in lap1]
           else:
              laps[w] = lap1
              [used.__setitem__(w2,w) for w2 in lap1]
    print
    print "Number of families",len(laps)
    
    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break
    
    print "Number of families left",len(laps)
    print Set([len(laps[w]) for w in ws if w in laps])

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())

    rats = []

    for w in laps:
        if len(laps[w])>2:
           vs = laps[w].keys()
           vs = [vs[j] for j in argsort([ellall[v].ab for v in vs])]
           evs = [ellall[v].v for v in vs]
           eabs = [ellall[v].ab for v in vs]
           rats1 = divz(eabs[1:],eabs[:-1])
           rats = concatenate([rats,rats1])

#    print
#    print sort(rats)
#    print
    qrats = quartiles(rats)
    print "Quartiles of ab ratios=",qrats
    efac = qrats[0]
#    efac = 1.23104806

    dups = {}
    newp = {}
    purge = {}

    for p in ws[:len(ws)]:
#    for p in []:
        if p in laps.keys():
           w2 = laps[p].keys()
           order = argsort([conall[w3].area for w3 in w2])
           w2 = [w2[j] for j in order] + [p]
           for j,w in enumerate(w2[:-1]):
#           w2 = [w2[j] for j in order]
#           for j,w in enumerate(w2):
               for v in w2[j+1:]:
                   vu = v
#                   if v in dups:
#                      vus = [v+"."+str(n+1) for n in range(dups[v])]
#                      vux = array([ellall[vu].mx for vu in vus])
#                      vuy = array([ellall[vu].my for vu in vus])
#                      vur = sqrt(vux**2+vuy**2)
#                      vu = vus[argmin(vur)]
#                   else:
#                      vu = v
                   interior = ellall[w].contour.multiply(efac).inside(conall[vu].x,conall[vu].y)
#                   if sum(interior) > 0.98*len(conall[vu].x):
#                      pass
                   if 12 < sum(interior) < len(conall[vu].x):
                      ix,iy = compress(interior, [conall[vu].x, conall[vu].y], 1)
                      iz = ones(iy.shape)
                      ellnew = Ellipse(ix,iy,iz,ns=0,verb=0)
                      areanew = pi*ellnew.a*ellnew.b
                      areaold = pi*ellall[w].a*ellall[w].b
#                      if areanew < areaold:
#                         purge.setdefault(w)
#                         print "here!!!!!"
#                         print v,w,sum(interior),len(interior),len(conall[vu].x)
#                         print ellall[v].mx,ellall[v].my,ellall[v].ab
#                         print ellall[w].mx,ellall[w].my,ellall[w].ab
#                         print ellnew.mx,ellnew.my,ellnew.ab
#                         print areanew,areaold
#                         print "here!!!!!"
                      if 1:
                         intnew = ellnew.contour.inside(conall[w].x,conall[w].y)
                         dx = (ellnew.mx-ellall[w].mx)
                         dy = (ellnew.my-ellall[w].my)
                         dr = sqrt(dx**2+dy**2)
                         if sum(interior) < len(conall[vu].x) and dr<0.1*ellall[w].ab:
                            interior = ellnew.contour.inside(ellall[w].ellx,ellall[w].elly)
                            if 0:
                               print
                               print "new ellipse!",vu,w,conall[w].mx,conall[w].my
                               print len(conall[vu].x),len(ix)
                               print ellall[vu].mx,ellall[vu].my,ellall[vu].ab
                               print ellnew.mx,ellnew.my,ellnew.ab
                               print sum(interior), len(interior),
                               print sum(1.*interior)/len(interior)
                            if sum(1.*interior) > 0.5*len(interior):
                                ellall[vu] = ellnew
                            else:
                                if v not in dups: dups[v] = 0
                                dups[v] += 1
                                v2 = v+"."+str(dups[v])
                                print "Adding",v2,dx,dy,dr
                                conall[v2] = copy.copy(conall[vu])
                                ellall[v2] = ellnew
                                if v2 not in ws:
                                   ws.append(v2)
                                if v2 not in laps:
                                   laps[v2] = {}
                                laps[v2].setdefault(w)
                                if w in laps[used[w]]:
                                   laps[used[w]].pop(w)
                                used[w] = v2

    for w in laps.keys():
        if not laps[w]: laps.pop(w)
    print "Number of families left",len(laps)

    print Set([len(laps[w]) for w in ws if w in laps])
    for w,n in dups.items():
        for k in range(n):
            w2 = w+"."+str(k+1)
            if w2 not in ws: ws.append(w2)

    print dups
    for w,n in dups.items():
        if w in laps:
           for k in range(n):
               w2 = w+"."+str(k+1)
               for v in laps[w2]:
                   if v in laps[w]:
                      laps[w].pop(v)
                      if not laps[w]: laps.pop(w)

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())

#    for l in sp:
#        if sp[l] > 1:
#           print l,sp[l],ellall[l].mx,ellall[l].my,ellall[l].ab
#           for k in laps:
#               if l in laps[k]:
#                  print k,len(laps[k]),ellall[k].mx,ellall[k].my,ellall[k].ab

#    exit()
#    for w in purge:
#        ws.remove(w)
#        for k in laps:
#            if w in laps[k]: laps[k].pop(w)

    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break

    for w in laps.keys():
        if not laps[w]: laps.pop(w)
    print Set([len(laps[w]) for w in ws if w in laps])
    print "Number of families left",len(laps)

    sp = {}
    [sp.__setitem__(l,sp.get(l,0)+1) for k in laps for l in laps[k]]
    print Set(sp.values())
    print Set([len(laps[w]) for w in ws if w in laps])
#    for w in laps:
#        if len(laps[w]) == 0:
#           print ellall[w].mx, ellall[w].my, ellall[w].a, ellall[w].b, ellall[w].theta
#    exit()

    inconseq = {}
    print "Refit ellipses?"
    if 0:
       for w in ws:
           if w in laps:
              interior = zeros(conall[w].x.shape,float)
#              print w,conall[w].mx,conall[w].my,conall[w].r,ellall[w].ab,conall[w].area
              for v in laps[w]:
                  dr = sqrt((conall[v].mx-conall[w].mx)**2+(conall[v].my-conall[w].my)**2)
#                  print v,conall[v].mx,conall[v].my,conall[v].r,ellall[v].ab,dr,
                  if (2*conall[v].r+dr) > 0.5*conall[w].r and conall[v].r < 0.25*conall[w].r:
#                  if dr > conall[w].r:
                     interior1 = conall[v].inside(conall[w].x,conall[w].y)
#                     print w,v,sum(interior1),
                     interior1 = conall[v].multiply(2).inside(conall[w].x,conall[w].y)
#                     print sum(interior1),
                     interior = interior + interior1
#                     print sum(interior),
#                  print
              ix,iy = compress(equal(interior,0), [conall[w].x, conall[w].y], 1)
#              print w,len(conall[w].x),len(ix),
              if not len(ix):
                 print "kill!",conall[w].mx,conall[w].my
                 inconseq.setdefault(w)
                 if w in laps:
                    [inconseq.setdefault(w3) for w3 in laps[w]]
              elif len(ix)!=len(conall[w].x):
                 iz = ones(iy.shape)
                 ellall[w] = Ellipse(ix,iy,iz,ns=0,verb=0)
#              print

    ws = [w for w in ws if w not in inconseq]

    for w in ws:
        xpixels, ypixels, zpixels = ellall[w].contour.pixels(data)
        wpixels = divz(zpixels*greater(zpixels,0),sqrt((getsqrt(zpixels)**2+sigma**2)))
        upixels = greater(wpixels,0)
        xpixels,ypixels,wpixels,zpixels = compress(upixels, \
                [xpixels,ypixels,wpixels,zpixels], 1)
        ellall[w].counts = sum(zpixels)

    cleaned = {}
    inconseq = {}
    watchout = {}

    for j,w in enumerate(ws):
        cleaning = True
#        cleaning = False
        while cleaning:
           cleaning = False
           if w in laps and len(laps[w]):
              xs,ys,fs = 0.0,0.0,0.0
              xl,yl,fl = [],[],[]
              for v in [w]+laps[w].keys():
                  dr = sqrt((ellall[v].mx-ellall[w].mx)**2+(ellall[v].my-ellall[w].my)**2)
                  f = sqrt(ellall[v].v*ellall[v].ab)*exp(-0.5*dr**2/ellall[w].ab**2)
                  fs += f
                  xs += ellall[v].mx*f
                  ys += ellall[v].my*f
                  xl.append(ellall[v].mx)
                  yl.append(ellall[v].my)
                  fl.append(f)
              xc,yc = divz(xs,fs), divz(ys,fs)
              xl,yl,fl = asarray([xl,yl,fl])
#              xc = median(xl)
#              yc = median(yl)
              xs,ys,fs = 0.0,0.0,0.0
              xl,yl,fl = [],[],[]
              for v in [w]+laps[w].keys():
                  dr = sqrt((ellall[v].mx-ellall[w].mx)**2+(ellall[v].my-ellall[w].my)**2)
                  f = (ellall[v].v*ellall[v].ab)*exp(-0.5*dr**2/ellall[w].ab**2)
                  fs += f
                  xs += (ellall[v].mx-xc)**2*f
                  ys += (ellall[v].my-yc)**2*f
                  xl.append(ellall[v].mx-xc)
                  yl.append(ellall[v].my-yc)
                  fl.append(f)
              xr,yr,rr = sqrt(divz(xs,fs)), sqrt(divz(ys,fs)), sqrt(divz(xs+ys,fs))
              xl,yl,fl = asarray([xl,yl,fl])
#              xr = 1.49*median(abs(xl))
#              yr = 1.49*median(abs(yl))
#              rr = 1.49*median(xl**2+yl**2)
              print "%20s %4d %10.3f %9.3f %10.3f %9.3f %9.3f" % (w,len(laps[w]),xc,xr,yc,yr,rr),
              bad = 0
              purge = {}
              purged = {}
              sf = 0.0
              killoff = False
              wl = laps[w].keys()
              wl = [wl[j2] for j2 in argsort([ellall[w1].ab for w1 in wl])[::-1]]
              for v in wl:
                  dx = (ellall[v].mx - xc)
                  dy = (ellall[v].my - yc)
                  dr = sqrt(dx**2+dy**2)/max([ellall[v].ab,2.0])
#                  df = divz(ellall[v].counts,cint)
#                  dc = divz(ellall[v].counts,divz(cint,care))
#                  nokeep = (dr>1.0) and (df>0.25 or (dc and dc>1.0))
                  nokeep = (dr>1.0)
                  if nokeep:
#                     if df > 1.00: killoff = True
                     purge.setdefault(v)
                     bad += 1
                     cleaning = True
              cint = 0.0
              care = 0.0
              wl = [v for v in wl if v not in purge]
              for v in wl:
                  dx = (ellall[v].mx - xc)
                  dy = (ellall[v].my - yc)
                  if sqrt(dx**2+dy**2)<=max([2*rr,4]):
                      cint = max([0.0,ellall[w].counts - ellall[v].counts])
                      care = max([0.0,ellall[w].area - ellall[v].area])
                      break
              print cint,
#              for v in purge:
#                  print v,ellall[v].mx, ellall[v].my, ellall[v].ab
              if cint <= 0.0: killoff = True
#              purge2 = {}
#              laps2 = {}
#              [laps2.setdefault(w3) for w2 in laps for w3 in laps[w2] if w2 != w]
#              [purge2.setdefault(v) for v in purge if v in laps2]
#              pcf = lambda v: 1.*sum(ellall[w].contour.inside(ellall[v].x,ellall[v].y))/len(ellall[v].x)
#              print [pcf(v) for v in purge2],
#              pc = sum([pcf(v)*ellall[v].counts for v in purge2])/ellall[w].counts
#              if pc > 4.00: killoff = True
#              print pc
#              for v in purge2:
#                  print v,ellall[v].mx, ellall[v].my, ellall[v].ab, ellall[v].counts
#              print
              while purge:
                 vs = purge.keys()
                 vs = [vs[j2] for j2 in argsort([ellall[v].area for v in vs])[::-1]]
                 if len(vs) > 1 and vs[0] not in laps:
                    laps[vs[0]] = {}
                    for v in vs[1:]:
                       laps[vs[0]].setdefault(v)
                    cleaning = True
#                     print conall[v].mx,conall[v].my,conall[v].r
#                     if purge[v] and v not in laps:
#                        killoff = True
#                        lap1 = GetOverlaps(w,ws[j+1:],conall,ellall)
#                        laps.setdefault(v,lap1)
#                        [inconseq.setdefault(v2) for v2 in laps if v in laps[v2]]
                 for v in vs:
                     laps[w].pop(v)
                     purge.pop(v)
                     purged.setdefault(v)
                     if v in laps:
                        for v2 in laps[v]:
                            if v2 in laps[w]:
#                               print conall[v2].mx,conall[v2].my,conall[v2].r
                               purge.setdefault(v2)
              if purged:
                 for v in [w]+laps[w].keys():
                     if v not in watchout: watchout[v] = {}
                     for v2 in purged:
                         watchout[v].setdefault(v2)
#                 for w2 in [w]+laps[w].keys():
#                 for w2 in []:
#                     dx = (conall[w2].mx - xc)
#                     dy = (conall[w2].my - yc)
#                     if w2 not in purged and sqrt(dx**2+dy**2)<=max([2*rr,4]):
#                        interior = zeros(conall[w2].x.shape,float)
#                        print
#                        for v in purged:
#                            interior1 = conall[v].inside(conall[w2].x,conall[w2].y)
#                            print w,v,sum(interior1),
#                            interior1 = conall[v].multiply(2).inside(conall[w2].x,conall[w2].y)
#                            print sum(interior1),
#                            interior = interior + interior1
#                            print sum(interior)
#                        ix,iy = compress(equal(interior,0), [conall[w2].x, conall[w2].y], 1)
#                        print len(conall[w2].x),len(ix)
#                        if len(ix)!=len(conall[w2].x):
#                           iz = ones(iy.shape)
#                           ellall[w2] = Ellipse(ix,iy,iz,ns=0,verb=0)
#                        elif not len(ix):
#                           inconseq.setdefault(w2)
#                           if w2 in laps:
#                              [inconseq.setdefault(w3) for w3 in laps[w2]]
              if killoff:
                 inconseq.setdefault(w)
                 [inconseq.setdefault(w2) for w2 in laps[w]]
              if bad: cleaned[w] = True
              print bad

    print "Refit ellipses?"
    if 0:
       for w in watchout:
           interior = zeros(conall[w].x.shape,float)
           for v in watchout[w]:
               interior1 = conall[v].inside(conall[w].x,conall[w].y)
#               print w,v,sum(interior1),
               interior1 = conall[v].multiply(2).inside(conall[w].x,conall[w].y)
#               print sum(interior1),
               interior = interior + interior1
#               print sum(interior)
           ix,iy = compress(equal(interior,0), [conall[w].x, conall[w].y], 1)
           print w,len(conall[w].x),len(ix),
           if not len(ix):
              print "kill!",conall[w].mx,conall[w].my
              inconseq.setdefault(w)
              if w in laps:
                 [inconseq.setdefault(w3) for w3 in laps[w]]
           elif len(ix)!=len(conall[w].x):
              iz = ones(iy.shape)
              ellall[w] = Ellipse(ix,iy,iz,ns=0,verb=0)
           print

    print "Number cleaned=",len(cleaned)
    print "Number inconseq=",len(inconseq)

    [laps.pop(w) for w in inconseq if w in laps]

    print
    print len(laps)
    mabs = {}
    for j,w in enumerate(ws):
        if w in laps and len(laps[w]) > 2:
           wabs = array([ellall[k].ab for k in laps[w]])
           wvs = array([ellall[k].v for k in laps[w]])
           wabs,wvs = take([wabs,wvs],argsort(wabs),1)
           warea = pi*wabs**2
           wf = add.accumulate( warea*wvs)
           wr = sqrt(divz(sum(wf*wabs**2),sum(wf)))
           if wr > 0: mabs[w] = wr

    wrs = sort(array([mabs[w] for w in mabs]))
    print "wrs"
    print "quartiles(wrs)",quartiles(wrs)
    print "extrema(wrs)",extrema(wrs)
    ars0 = arange(-1.0,2.0,0.05,float32)
    nrs = diff(1.0*searchsorted(log10(wrs),ars0))
    ars = diff(ars0)/2.0 + ars0[:-1]
    crs = add.accumulate(nrs)/sum(nrs)
    nrs,ars,crs = compress(nrs,[nrs,ars,crs],1)

    if less(diff(nrs),0).any():
       see = 10.0**ars[nonzero(less(diff(nrs),0))[0]]
       print "Radius=",see

    print "Prune families from",len(laps)
    popping = True
    pn = 0
    pruned = {}
    while popping:
       popping = False
       pn += 1
       if pn and not (pn%100):
          sys.stdout.write(".")
          sys.stdout.flush()
       for w in ws:
           if w in laps:
              for w2 in laps[w].keys():
                  if w!=w2 and w2 not in pruned and w2 in laps:
                     [laps[w].setdefault(w3) for w3 in laps[w2]]
                     [used.__setitem__(w3,w) for w3 in laps[w2]]
                     pruned.setdefault(w2)
                     laps.pop(w2)
                     popping = True
                     break
           if popping: break

    print "Number of families left",len(laps)

    parents = {}

    nl = len(laps)
    vn = 10.0**(int(log10(nl+1)+1))
    n = len(ws)
    value = 0
    used = {}
    drawings = []
    colors = []
    fills = []
    for j in range(n):
        if ws[j] not in used and ws[j] not in inconseq:
           uw = [ws[j]]
           luw,nuw = 0,1
           while nuw!=luw:
              luw = len(uw)
              for w2 in uw:
                 if w2 in laps:
                    [uw.append(w) for w in laps[w2] if w not in uw]
              nuw = len(uw)
           uw = [w for w in uw if w not in used]
           if uw:
              value = value+1
#              print value,len(uw),len(used)
#              sys.stdout.flush()
              order = argsort([conall[w].area for w in uw])[::-1]
              uw = [uw[j2] for j2 in order]
              parents[uw[0]] = {}
              for w in uw:
                  parents[uw[0]][w] = (conall[w],ellall[w])
                  used.setdefault(w)
              uw = [w for w in uw if ellall[w].ab > 1]
              for w in uw:
                  if w == uw[0] or w == uw[-1]:
                     if efit:
                        color = conall[w].v
                        x, y = ellall[w].ellx, ellall[w].elly
                        x = clip(x,0,naxis1-1)
                        y = clip(y,0,naxis2-1)
#                        print w,extrema(x), extrema(y)
                        color = value/vn
                        color = conall[w].v
                        color = -999999
                        segments = asarray([[(x[j],x[j+1]),(y[j],y[j+1])] for j in range(len(x)-1)])
                     else:
                        color = value/vn
#                        color = conall[w].v
                        color = -999999
                        segments = conall[w].contour
#                        segments = conall[w].multiply(2.0).contour
                     drawings.append(segments)
                     colors.append(color)
                     if w == uw[-1]:
                        if efit:
#                           fill = value/vn
#                           fill = conall[w].v
                           fill = ellall[w].counts/ellall[w].area
                        else:
                           fill = ellall[w].counts/ellall[w].area
                     else:
                         fill = 0 
                     fills.append(fill)

    oframe = zeros((naxis2,naxis1),float64)
    if len(drawings):
       print
       print len(drawings)
       sys.stdout.flush()
       oframe = VTKDrawContour(zeros(oframe.shape,float64),drawings,color=colors,fill=fills,verb=1).astype(float32)
    print
    sys.stdout.flush()
    if efit: oimage = "efamilies%04d.fits" % (n1)
    else: oimage = "families%04d.fits" % (n1)
    qdump(oimage,oframe)
    zdb = "families%04d.zodb" % (n1)
    if efit: WriteContours(zdb,parents)

