
from basics import *
from numpy.core.numerictypes import *
import numpy
from sets import Set
from Expector import *
from LDP import *
from FITS import *
from Multipack import *
from Mestimators import *
import os

pkeys = ["subsec1", "subsec1m", "subsec1r", "subsec1b", "subsec1n",
        "subsec1X", "subsec1y", "subsec1w", "subsec1gh", "fitted1"]
dkeys = ["ds","ms","rs","bs","ns",
         "xs","ys","ws","ps","fs"]
ldkeys = dkeys + ["dws","ts","qs"]


def OneIndex(dfile,verb=0):
    d = FITS(dfile,readonly=1)
    objects = []
    while not len(objects) or objects[-1] != "N/A":
       objects.append(d["OBJ%04d" % (len(objects)+1)].strip())
    d.close()
    objects = objects[:-1]
    if verb: print "Read %d object names from %s" % (len(objects),dfile)
    return objects

def ArrayToDict(X,t):
    a = {}
    for jkey,dkey in enumerate(dkeys): a[dkey] = X[jkey]
    a["dws"] = Derivative(a["xs"],a["ws"])
    a["ts"] = t*ones(a["xs"].shape,float64)
    return a

def Derivative(x,w):
    wa = sort(list(Set(w.ravel())))
    xa = array([median(compress(equal(w.ravel(),w1),x.ravel())) for w1 in wa])
    xa,wa = take([xa,wa],argsort(xa),1)
    xtck = splrep(xa,log(wa),s=0,k=3)
    dlwdx = zeros(x.ravel().shape,float64)
    order = argsort(x.ravel())
    put(dlwdx,order,splev(take(x.ravel(),order),xtck,1))
    dws = reshape(dlwdx*w.ravel(),x.shape)
    return abs(dws)

def ReadOneFile(dfile,objno,getdata=1,verb=0):
    if verb: print "Reading from %s[%d]" %(dfile,objno)
    qs = None
    if "odc" in dfile and getdata:
       qfile = dfile.replace("odc","oqc")
       if os.path.exists(qfile):
          d = FITS(qfile,readonly=1)
          qd = d.data()
          qw = qd[3][objno-1]
          qq = qd[5][objno-1]
          qf = qd[6][objno-1]
          d.close()
          qs = (qw,qq,qf)
    d = FITS(dfile,readonly=1)
    etime = d["exptime"]
    if d["nshuffle"]:
      ccdtype = d["ccd-type"]
      epoch = d["epoch"]
      if ccdtype == "E2V": zpa = 43.80
      elif epoch < 2005.5: zpa=223.53
      else: zpa=133.55
      pa = -(d["rotangle"]-zpa) * pi/180
      nr = d["nod-ra"]/0.2
      nd = d["nod-dec"]/0.2
      nsh = sqrt(nd**2 + nr**2)/2.
      if verb: print nr,nd,pa,nsh,
    else: nsh = 0
    if nsh: nf = 2
    else: nf = 1
    hap = d["hap"]
    corners = []
    hdu = nf*(objno-1)+2
    if verb: print hdu,
    d.goto(hdu)
    slitobj = d["slitobj"]
    if verb: print slitobj,
    rchi2 = d["rchi2"]
    if verb: print "%.4f" % (rchi2),
    pm1 = []
    while d["pm1"+str(len(pm1)+1)] != "N/A": pm1.append(d["pm1"+str(len(pm1)+1)])
    if verb: print pm1
    pm1 = eval("["+"".join(pm1)+"]")[:1]
    pm2 = []
    while d["pm2"+str(len(pm2)+1)] != "N/A": pm2.append(d["pm2"+str(len(pm2)+1)])
    if verb: print pm2
    pm2 = eval("["+"".join(pm2)+"]")[:1]
    pmh = []
    while d["pmh"+str(len(pmh)+1)] != "N/A": pmh.append(d["pmh"+str(len(pmh)+1)])
    if verb: print pmh
    pmh = eval("["+"".join(pmh)+"]")
    if getdata:
       corners.append(eval(d["corners"]))
       adata = d.data()
       if nsh:
          d.goto(hdu+1)
          slitobj = d["slitobj"]
          if verb: print slitobj
          corners.append(eval(d["corners"]))
          bdata = d.data()
    d.close()
#    qdump("/tmp/a.fits",adata[0])
#    qdump("/tmp/b.fits",bdata[0])
#    exit()
    o = {"t":etime,"h":hap,"nf":nf,"nsh":nsh,"rchi2":rchi2,"qs":qs}
    o["pm1"] = asarray(pm1)
    o["pm2"] = asarray(pm2)
    o["pmh"] = asarray(pmh)
    if getdata:
       o["corners"] = corners
       o["a"] = adata
       if nsh:
          o["b"] = bdata
    return o

def ReadOneObject(object,robjects,objno):
    print object
    robject = robjects[object]
    print robject
#    objno = [1+dobjects[dfile].index(object) for dfile in robject]
    print objno
    data = {}
    for r,o in zip(robject,objno):
       data[r] = ReadOneFile(r,o,getdata=1,verb=0)
    for r in data:
       d = data[r]
       d["a"] = ArrayToDict(d["a"],d["t"])
       if d["nsh"]: d["b"] = ArrayToDict(d["b"],d["t"])
       if not d["qs"]:
           d["haveqe"] = False
           d["a"]["qs"] = ones(d["a"]["ds"].shape,float32)
           if d["nsh"]: d["b"]["qs"] = ones(d["a"]["ds"].shape,float32)
       else:
           d["haveqe"] = True
           ws = d["a"]["ws"]
           if d["nsh"]:
             wsb = d["b"]["ws"]
           qw,qq,qf = d["qs"]
           qf = qf
           qd = {}
           [qd.setdefault(qw1,(qq1,qf1)) for qw1,qq1,qf1 in zip(qw,qq,qf)]
           ws0 = ws[0]
           qs0 = array([(ws1 in qd and qd[ws1][0] or 0.0) for ws1 in ws0])
           qf0 = array([(ws1 in qd and qd[ws1][1] or 0.0) for ws1 in ws0])
           qs = asarray(len(ws)*[qs0])
           qf = asarray(len(ws)*[qf0])
           d["a"]["qs"] = qf*qs
           if d["nsh"]:
               qsb = asarray(len(wsb)*[qs0])
               qfb = asarray(len(wsb)*[qf0])
               d["b"]["qs"] = qfb*qsb
           root = os.path.basename(r).split(".")[0]
#           [qdump("/tmp/%s_%sa.fits" % (root,dkey),d["a"][dkey]) for dkey in dkeys]
#           [qdump("/tmp/%s_%sb.fits" % (root,dkey),d["b"][dkey]) for dkey in dkeys]
#           exit()
#           for k in ["ds","ms","rs","bs","fs","ns"]:
#              d["a"][k] = divz(d["a"][k],qs)
#              if d["nsh"]:
#                  d["b"][k] = divz(d["b"][k],qsb)
    return data

WBinsDict = {}

def GetWBins(nbin=1.0,minbin=0.0,w1=0.0,w2=0.0):
    if (nbin,minbin,w1,w2) not in WBinsDict:
       mw = arange(3800,11.0e3,5.0,float32)
       if w1: mw = concatenate([mw,[w1]])
       if w2: mw = concatenate([mw,[w2]])
       mw = sort(list(Set(mw)))
       wxbasis = legendre((mw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       mx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
#       for (a,b) in zip(mw,mx): print a,b
       if w1: mx0 = mx[searchsorted(mw,w1)]
       else: mx0 = mx[0]
       mx = mx - mx0
       lmw = log(mw)
       mxtck = splrep(mx,lmw,s=0,k=1)
       mdwdx = splev(mx,mxtck,1)*mw
#       mdwdx = clip(mdwdx,minbin,1e3)
       if w1: bin0 = w1
       else: bin0 = mw[0]+mdwdx[0]/2.0
       if w2: wmax = w2
       else: wmax = LDPDISP["lambda"][-1]+nbin*mdwdx[-1]/2.
       def getx(www,xxx=-1):
           wxbasis = legendre((www-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
           p = (LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0)
#           raw_input(str((p,xxx)))
           if xxx == -1: return p
           else: return p - xxx
       bins = [bin0]
       xins = [getx(bin0)]
#       print nbin
       while bins[-1] < wmax:
#          print xins
#          print bins,wmax
          new = fsolve(getx,[bins[-1]],args=(xins[-1]+nbin,))
#          raw_input(str(new))
          xins.append(xins[-1]+nbin)
          bins.append(new)
       wbins = array(bins)
       if w2: wbins = compress(less_equal(wbins,w2),wbins)
       xbins = getx(wbins)
       dbins = nbin*splev(xbins,mxtck,1)*wbins
#       dbins = nbin * clip(dbins,minbin,1e3)
       WBinsDict[(nbin,minbin)] = wbins,dbins
    return WBinsDict[(nbin,minbin)]


def oldGetWBins(nbin=1.0,minbin=0.0,w1=0.0,w2=0.0):
    if (nbin,minbin,w1,w2) not in WBinsDict:
       mw = LDPDISP["lambda"]
       if w1: mw = concatenate([mw,[w1]])
       if w2: mw = concatenate([mw,[w2]])
       mw = sort(list(Set(mw)))
       wxbasis = legendre((mw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       mx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
       mx,mw = take([mx,mw],argsort(mx),1)
       if w1:
          mx0 = mx[searchsorted(mw,w1)]
       else:
          mx0 = mx[0]
       mx = mx - mx0
       lmw = log(mw)
       mxtck = splrep(mx,lmw,s=0,k=3)
       mdwdx = splev(mx,mxtck,1)*mw
       mdwdx = clip(mdwdx,minbin,1e3)
       if w1: bin0 = w1
       else: bin0 = mw[0]+mdwdx[0]/2.0
       if w2: wmax = w2+nbin*mdwdx[searchsorted(mw,w2)]/2.
       else: wmax = LDPDISP["lambda"][-1]+nbin*mdwdx[-1]/2
       bins = [bin0]
       while bins[-1] < wmax:
          bw0 = 1e38
          bw = bins[-1]
          co = 0
          while abs(bw-bw0)/bw > 1e-6:
             bw0 = bw
             co += 1
             if len(bins) > 1:
                wxbasis = legendre((array([bins[-1],bw])-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
                bx = mean(LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0)
             else:
                wxbasis = legendre((bw-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
                bx = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0
             dwdxb = splev(bx,mxtck,1)*bw
             dwdxb = clip(dwdxb,minbin,1e3)
             bw = bins[-1]+nbin*dwdxb/2.
             print co,bw,bw0,dwdxb
             if bw > 9000: raw_input(str(bins))
          bins.append( bins[-1]+nbin*dwdxb)
       wbins = array(bins)
       if w2: wbins = compress(less_equal(wbins,w2),wbins)
       wxbasis = legendre((wbins-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
       xbins = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ - mx0
       dbins = splev(xbins,mxtck,1)*wbins
       dbins = nbin * clip(dbins,minbin,1e3)
       WBinsDict[(nbin,minbin)] = wbins,dbins
    return WBinsDict[(nbin,minbin)]

def SingleBin(w,dw,d,h,nsh=1,t=0,M=L2,pw=1,scut=5.0,haveqe=1):
    if scut > 100:
       adw = [1.0]
    else:
       adw = [1.0]
    for jdw,ndw in enumerate(adw):
       sws = sort(list(Set(d["ws"])))
       keep = less(abs(d["ws"]-w),ndw*dw/2.)
#       print w,ndw*dw/2.,len(keep.ravel()),sum(keep.ravel()),
       if haveqe: keep = keep*greater(d["qs"],0)
       if not sum(keep.ravel()):
           print w,ndw*dw/2.,len(keep.ravel()),
           print sum(keep.ravel()),
           print "not enough!"
#           print sws
           return asarray(6*[0.0])
       e0 = {}
       for dkey in ldkeys: e0[dkey] = compress(keep.ravel(),d[dkey].ravel())
       if "yc" in e0:
          keep = less_equal(abs(e0["yc"]),h+0.5)
       else:
          yc = e0["ys"].ravel()[argmax(e0["ps"])]
          keep = less_equal(abs(e0["ys"]-yc),h+0.5)
       e = {}
       for dkey in ldkeys:
          e[dkey] = compress(keep.ravel(),e0[dkey].ravel())
       dt = e["ts"]
       dd = abs(e["dws"])
       y = e["ys"] - yc
       p = e["ps"]
       f = e["fs"]
       s = e["ds"]
       n = e["ns"]
       qe = e["qs"]
       r0 = (s-f)/n
       er = M(r0/scut,w=1)
       k = 1/(dt*dd*p)
       s = s*k/qe*1e20
       n = n*k/qe*1e20
       v = n**2
       q = divz(1.0,v)
       spec1 = divz(sum(er*q*s),sum(er*q))
       nois1 = divz(1,getsqrt(divz(sum(er*q*(divz(1,v))),sum(er*q))))
       if jdw == 0:
          Me = less(abs(r0),scut)
       else:
#          nfac = bwt((s-spec2/k)/n)[1]
          nfac = 1.0
#          print nfac
#          Me = less(abs(r0),scut)
          Me = Fair((s-spec2/k)/n/nfac,w=1)
#          Me = Huber2((s-spec2/k)/n/nfac,w=1)
#          Me = L1((s-spec2/k)/n/nfac,w=1)
#          Me = L2((s-spec2/k)/n/nfac,w=1)
#          Me = M(r0/scut,w=1)
       qM = power(p,pw)*q
       nqM = mean(compress(qM,qM))
       if nqM: qM = qM/nqM
#       qM = q
       spec2 = divz(sum(qM*s),sum(qM))
       nois2 = sqrt(divz(1,len(nonzero(qM))*divz(sum(qM/v),sum(qM))))
       scat = divz(sum(q*(s-spec2)**2),sum(q))
       scat = sqrt(scat)
    if 5000 < w < 8000 and nois2 == 0:
       print "here",w,dw,e["ys"].shape
       print e["ys"]
    apkeep = less_equal(abs(e["ys"]),h+0.5)
    A = 2.0*h/1e20
    R = A*t
    if nsh != 0: R = 2*R
    spec2 = spec2 * R
    nois2 = nois2 * R / sqrt(2.0*h)
    spec1 = spec1 * R
    nois1 = nois1 * R / sqrt(2.0*h)
    eds,ens,edws,er0 = compress(apkeep,[e["ds"],e["ns"],e["dws"],r0],1)
#    eds,ens,edws,er0 = e["ds"],e["ns"],e["dws"],r0
    eds = eds/edws
    ens = ens/edws
#    wt = divz(1,ens**2*(1+er0**2))
    wt = divz(1,ens**2)
    spec0 = divz(sum(wt*eds),sum(wt))*A*1e20
    nois0 = sqrt(divz(sum(wt*ens**2),sum(wt)*len(nonzero(wt))))*A*1e20
    pkg = asarray([spec0,nois0,spec1,nois1,spec2,nois2])
    return pkg

def FitOneObject(data,nbin=1,output="Spectra",name=None,inter=0,scut=5.0,w1=0,w2=0,hap=0,momfix=0):
    wbins,dbins = GetWBins(1,minbin=1.0,w1=w1,w2=w2)
    w1a = w1
    w2a = w2
    ro = data.keys()
    ro.sort()
#    ro = [r for r in ro if data[r]["rchi2"] < 2]
    nsh = data[ro[0]]["nsh"]
    haveqe = data[ro[0]]["haveqe"]
    unsh = [0,1][nsh!=0]
    if not hap: h = mean([data[r]["h"] for r in ro])
    else: h = hap

#    if inter: pdev = "/xs"
#    else: pdev = "%s/%s.ps/vcps"

#    Mp = lambda r,w: 1.0/(1+r**2)
    if haveqe and scut < 100:
       Mp = lambda r,w: 1.0
#       Mp = lambda r,w: 1.0/(1+r**2)
       Mp = Fair
       Mps = [Mp,Mp,Mp,Mp,Mp,Mp]
       pbins = [2.0,2.0,1.5,1.5,1.0,1.0]
    else:
       Mp = lambda r,w: 1.0
       Mps = [Mp,Mp,Mp,Mp]
       pbins = [1.0,1.0,1.0,1.0]
    npit = len(Mps)
    for pit,pbin,pMp in zip(range(npit),pbins,Mps):
       for r in ro:
          for jab,ab in enumerate("ab"[:unsh+1]):
              ds = data[r][ab]["ds"]
              fs = data[r][ab]["fs"]
              ns = data[r][ab]["ns"]
              bs = data[r][ab]["bs"]
              qs = data[r][ab]["qs"]
              mbs = median(bs)
              rs = divz(ds-fs,ns)**2
              rrs = divz(ds-fs,scut*ns)**2
              brs = divz(bs-mbs,ns)**2
              wrs = 1.0*greater(rrs,1.0)
#              wrs = logical_or(greater(rrs,1.0),greater(brs,0.1))
#              qdump("/tmp/dbs.fits",divz(bs-mbs,ns))
#              [qdump("/tmp/%s.fits" % (key),data[r][ab][key]) for key in ldkeys]
#              exit()
              ls = 1.0*less(VTKConvolve(1.0*wrs,k=3,numret=1),0.25)
              if unsh and not momfix: ls = ls * greater(bs,0)
#              else: ls = ls * greater(ns,0)
              data[r][ab]["ls"] = ls

              ws = data[r][ab]["ws"].ravel()
              ys = data[r][ab]["ys"].ravel()
              if not pit:
                 m1b = legendre((ws-LDPCOEFCi)/LDPCOEFSi, len(data[r]["pm1"]) - 1)
                 dy1 = dot(data[r]["pm1"],m1b)
                 yc = ys - dy1
                 data[r][ab]["yc"] = reshape(yc, data[r][ab]["ys"].shape)
                 data[r][ab]["ys"] = data[r][ab]["yc"]

       tt = 0
       t = []
       if inter > 1 or (pit == npit-1):
          nx=2*(1+unsh)
          ny=(1+unsh)*len(ro)/nx + 1
#          if len(ro) > 4:
#             nx = nx + (1+unsh)
#             ny=(1+unsh)*len(ro)/nx+1
##             ny = 6
          if nx==ny==1:psize = 8
          else:
             asp = 0.75*ny/nx
             print nx,ny
             print asp
             psize=min([nx*4,12/asp])
             print psize
          if inter: pdev = PGPLOT_DEV
          else: pdev = "%s/%s_2D.png/png" % (output,name)
          opendev(pdev,nx=nx,ny=ny,size=psize,land=0,asp=0.75*ny/nx,ask=1)
          for jr,r in enumerate(ro):
             for jab,ab in enumerate("ab"[:unsh+1]):
                corners = data[r]["corners"][jab]
                ds = data[r][ab]["ds"]
                bs = data[r][ab]["bs"]
                fs = data[r][ab]["fs"]
                ps = data[r][ab]["ps"]
                ns = data[r][ab]["ns"]
                ws = data[r][ab]["ws"]
                xs = data[r][ab]["xs"]
                ys = data[r][ab]["ys"]
                ls = data[r][ab]["ls"]
                qs = data[r][ab]["qs"]
                ors = data[r][ab]["rs"]
                rs = divz(ds-fs,ns)
                gz = ps
                wrz = 1/(1+rs**2)
#                qdump("/tmp/ns%d%s.fits" % (jr,ab),ns)

                for js,us in enumerate([ds,ds-fs,ls*rs]):
                   if js == 0:
                      cpgsvp(0.1,0.95,0.65,0.90)
                   elif js == 1:
                      cpgsvp(0.1,0.95,0.40,0.65)
                   elif js == 2:
                      cpgsvp(0.1,0.95,0.15,0.40)
                   xmar = add(extrema(xs.ravel()),[-0.5,2])
                   ymar = add(extrema(ys.ravel()),[-0.5,-2.0])
                   newbox(xmar,ymar,ch=2.5,manvp=1,xlab=(js==2),ylab=2,era=(js==0))
                   if js == 0:
                     xfm = array([xs.ravel()[0]-0.5,1.0,0.0,ys.ravel()[0]-0.5,0.0,1.0])
                     subtitle("%s (%s)" % (r,ab),ch=2.5)
                     mds = 1.49*median(abs(compress(ps.ravel()>1e-4,us.ravel())))
                     if mds: a1,a2 = -5*mds,5*mds
                     else: a1,a2 = -100,100
#                     print a1,a2
                   elif js == 2:
                     mds = 1.49*median(abs(compress(ps.ravel()>1e-4,us.ravel())))
                     a1,a2 = -3,3
#                     if mds: a1,a2 = -5*mds,5*mds
#                     else: a1,a2 = -3,3

                   if pdev[-2:] == "ps": a1,a2=a2,a1

#                   mds = divz(sum(wrz.ravel()*gz.ravel()*ps.ravel()*ds.ravel()),sum(wrz.ravel()*gz.ravel()*ps.ravel()))
#                   mds = median(ns.ravel())

                   us = where(ns,us,-999)

                   cpggray(us,us.shape[1],us.shape[0],1,us.shape[1],1,ds.shape[0],a1,a2,xfm)

                   cpgsci(5)
                   cpgsls(4)
#                   cont = array([data[r]["pm1"]])
                   cont = array([0.0])
                   cpgcont(ys,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   cpgsci(7)
                   cpgsls(2)
#                   cont = array([max(ps.ravel())/2.0])
#                   cpgcont(ps,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cont = sqrt(array([data[r]["pm2"][:1]]))
                   cont = concatenate([-sqrt(array([data[r]["pm2"][:1]])),cont]) #+ data[r]["pm1"][0]
                   cpgcont(ys,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   cpgsci(2)
                   cpgsls(1)
                   cont = array([ data[r]["nsh"]])
                   cpgcont(abs(ys),ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   cpgsci(8)
                   cpgsls(4)
                   cont = array([w1,5000,6000,7000,8000,9000,1e4,w2])
                   cpgcont(ws,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                   cpgsci(1)
                   cpgsls(1)

                   if not js:
                      cpgsci(4)
                      cpgsls(1)
                      cont = array([scut])
                      cpgcont(abs(rs),ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                      cpgsci(1)
                      cpgsls(1)
                      cpgsci(3)
                      cpgsls(1)
                      cont = array([1.0])
                      cpgcont(ls,ps.shape[1],ps.shape[0],1,ps.shape[1],1,ps.shape[0],cont,-len(cont),xfm)
                      cpgsci(1)
                      cpgsls(1)


          curs = ["O",(0,0)]
          while inter and curs[0] != "g":
             if curs[0] in "qQI": os._exit(0)
             else: curs = getcurs(curs[1][0],curs[1][1])
          closedev()

       d = {}
       for r in ro:
          for ab in "ab"[:unsh+1]:
             t.append(data[r]["t"])
          d[r] = {}
          for dkey in data[r]["a"]:
             d[r][dkey] = []
             for ab in "ab"[:unsh+1]:
                d[r][dkey] = concatenate([d[r][dkey],data[r][ab][dkey].ravel()])
          print r, len(d[r]["ps"]),sum(1.*greater(d[r]["ps"],1e-3)),
          print sum(1.*greater(d[r]["ns"],0)),
          print sum(1.*greater(d[r]["ds"],-3*d[r]["ns"])),
          print sum(1.*greater(d[r]["ls"],0)),
          keep = 1.*not_equal(d[r]["ds"],0)
          keep = keep*greater(d[r]["ns"],0)
          keep = keep*d[r]["ls"]
          keep = keep*greater(d[r]["ds"],-3*d[r]["ns"])
          if haveqe: keep = keep*greater(d[r]["qs"],0)
          keep = keep*greater(d[r]["ps"],1e-4)
          if nsh:
             keep = keep*less(abs(d[r]["ys"]), data[r]["nsh"])
##          keep = keep*less(abs(d[r]["ys"]-data[r]["pm1"]),h+1)
          print "sum(keep)=",sum(keep)
          for dkey in ldkeys: d[r][dkey] = compress(keep,d[r][dkey])
          order = argsort(d[r]["ws"])
          for dkey in ldkeys: d[r][dkey] = take(d[r][dkey],order)
    
       tt = sum(t)
       A = 2.0*h
       print t
#       exit()

       w1b,w2b = extrema([extrema(data[r][ab]["ws"].ravel()) for r in ro for ab in "ab"[:unsh+1]])

       wbins,dbins = GetWBins(pbin*nbin,minbin=1.0,w1=w1,w2=w2)

       individ = {}
       norms = {}
       errors = {}
       scales = {}
       weights = {}
       if pit: pw = 2
       else: pw = 2
       for r in ro:
          individ[r] = transpose([SingleBin(wb,db,d[r],h,nsh=unsh,t=data[r]["t"],M=pMp,pw=pw,scut=scut,haveqe=haveqe) for wb,db in zip(wbins,dbins)])
#          print individ[r][4]
#          print individ[r][5]
          for x in range(3):
             individ[r][2*x] = individ[r][2*x]/((unsh+1)*data[r]["t"])
             individ[r][2*x+1] = individ[r][2*x+1]/((unsh+1)*data[r]["t"])
          wtr = divz(between(wbins,5000.0,8500.0),(1e20*individ[r][5])**2)*greater(individ[r][4],-3*individ[r][5])
          norms[r] = divz(sum(wtr*individ[r][4]), sum(wtr))
          errors[r] = divz(sum(wtr*divz(individ[r][4],individ[r][5])), sum(wtr))

       print [norms[r] for r in ro]
       print [errors[r] for r in ro]
       unorms = [norms[r] for r in ro]
       uerrors= [errors[r] for r in ro]
       gnorms = compress(greater(unorms,0),unorms)
       gerrors = compress(greater(uerrors,0),uerrors)
       if len(gnorms): mnorm = median(gnorms)
       else: mnorm = 0.0
       if len(gerrors): merror = median(gerrors)
       else: merror = 0.0
       for r in ro:
          newnorm = (norms[r] > median(gnorms)/4 and norms[r]) or 0.0
          newerror = (errors[r] > median(gerrors)/4 and errors[r]) or 0.0
          print r,norms[r],mnorm,divz(norms[r],mnorm),newnorm
          print r,errors[r],merror,divz(errors[r],merror),newerror
          norms[r] = newnorm
          errors[r] = newerror
          print
       nnorm = [norms[r] for r in ro]
       nerror = [errors[r] for r in ro]
       mnorm = mean(compress(nnorm,nnorm))
       merror = mean(compress(nerror,nerror))

       print [norms[r] for r in ro]
       sumweights = 0.0
       nro = 0
       tt = 0.0
       for r in ro:
          if mnorm:
             norms[r] = divz(mnorm,norms[r])
             if norms[r]:
                tt = tt + data[r]["t"]*(unsh+1)
                nro = nro + 1
          else:
             norms[r] = 1.0
          scales[r] = norms[r]*data[r]["t"]*(unsh+1)
          weights[r] = divz(1,sqrt(scales[r])*individ[r][5]**2)
          sumweights = sumweights + weights[r]
       print [norms[r] for r in ro]
       sumnorms = sum([norms[r] for r in ro])
       sum2norms = sum([norms[r]**2 for r in ro])
       sumweights = divz(sumweights, nro)
       for r in ro:
          weights[r] = divz(weights[r],sumweights)
#       print [median(weights[r]) for r in ro]
       print nro

       print A,tt
       sbins2 = zeros(wbins.shape,float32)
       ebins2 = zeros(wbins.shape,float32)
       bbins2 = zeros(wbins.shape,float32)

#       for x in range(3):
#          exec("sbins%d = divz(sum([(weights[r]*scales[r]*individ[r][%d]) for r in ro]),tt)" % (x,2*x))
#          exec("ebins%d = divz(sqrt(sum([weights[r]*(scales[r]*individ[r][%d])**2 for r in ro])),tt)" % (x,2*x+1))

       if pbin == 1.0 or 1:
         ds = []
         bs = []
         vs = []
         ns = []
         fs = []
         ps = []
         ws = []
         dws = []
         qs = []
         tx = None
         for r in ro:
            if mnorm and norms[r]:
               print r
               ws2 = []
               dws2 = []
               ds2 = []
               fs2 = []
               bs2 = []
               vs2 = []
               ns2 = []
               ps2 = []
               qs2 = []
               for ab in "ab"[:unsh+1]:
                  if "ys" in data[r][ab]:
                     keep = 1.0*less_equal(abs(data[r][ab]["ys"].ravel()),h)
                     if nsh: keep = keep*less(abs(data[r][ab]["ys"].ravel()), data[r]["nsh"])
                  else:
                     ws1 = data[r][ab]["ws"].ravel()
                     ys1 = data[r][ab]["ys"].ravel()
#                     wxbasis = legendre((ws1-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                     dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
                     m1 = data[r]["pm1"]
                     if len(shape(m1)) == 0: m1 = [m1]
#                     m1b = polynom(ws1/1e4,len(m1)-1)
#                     m1b = polynom(ws1/1e4,len(m1)-1)
                     m1b = legendre((ws1-LDPCOEFCi)/LDPCOEFSi,len(m1)-1)
                     dy1 = dot(m1,m1b)
                     yc1 = ys1 - dy1
                     keep = 1.0*less_equal(abs(yc1.ravel()),h)
                     if nsh: keep = keep*less(abs(yc1), data[r]["nsh"])
                  keep = keep * (data[r][ab]["ls"].ravel())
                  keep = keep * greater(data[r][ab]["ns"].ravel(),0.0)
                  keep = keep * greater(data[r][ab]["ws"].ravel(),0)
                  if haveqe: keep = keep * greater(data[r][ab]["qs"].ravel(),0)
#                  keep = keep * between(data[r][ab]["ws"].ravel(),w1,w2,inc=1)
                  keep = keep * greater(data[r][ab]["ps"].ravel(),1e-4)
                  keep = keep * greater(data[r][ab]["ds"].ravel(),-3* data[r][ab]["ns"].ravel())
                  keep = keep * not_equal(data[r][ab]["ds"].ravel(),0)
#                  keep = keep * not_equal(data[r][ab]["fs"].ravel(),0)
#                  keep = keep * less(abs(data[r][ab]["ds"].ravel()-data[r][ab]["fs"].ravel()),scut*data[r][ab]["ns"].ravel())
                  ds1 = compress(keep,data[r][ab]["ds"].ravel())
                  qs1 = compress(keep,data[r][ab]["qs"].ravel())
                  bs1 = compress(keep,data[r][ab]["bs"].ravel())
                  ns1 = compress(keep,data[r][ab]["ns"].ravel())
                  ws1 = compress(keep,data[r][ab]["ws"].ravel())
                  dws1 = compress(keep,data[r][ab]["dws"].ravel())
                  fs1 = compress(keep,data[r][ab]["fs"].ravel())
                  ps1 = compress(keep,data[r][ab]["ps"].ravel())
                  ms1 = pMp((ds1-fs1)/ns1,w=1)
                  ds1 = ds1*norms[r]/ps1/dws1/data[r]["t"]/qs1
                  fs1 = fs1*norms[r]/ps1/dws1/data[r]["t"]/qs1
#                  ns1 = ns1*sqrt(norms[r]/ps1/dws1/data[r]["t"]/qs1)
                  ns1 = ns1*norms[r]/ps1/dws1/data[r]["t"]/qs1
                  bs1 = bs1*norms[r]/dws1/data[r]["t"]/qs1
                  vs1 = ns1**2
                  ns1 = divz(ms1,ns1)
#                  ns1 = ns1 * sqrt(norms[r]/ps1/dws1/data[r]["t"]/qs1)
#                  ns1 = ns1*sqrt(data[r]["t"]*qs1*ps1*dws1/norms[r])
#                  ns1 = ps1*ns1*(data[r]["t"]/norms[r])/sqrt(data[r]["pm2"][0])
#                  ns1 = ps1*ns1*(data[r]["t"]/norms[r])/sqrt(data[r]["pm2"][0])
#                  ns1 = ps1*ns1*(data[r]["t"]/norms[r])
#                  ns1 = ps1*ns1
                  ns1 = ps1*ns1*(data[r]["t"]/norms[r])
                  ws1,dws1,ds1,bs1,vs1,ns1,fs1,ps1,qs1 = take([ws1,dws1,ds1,bs1,vs1,ns1,fs1,ps1,qs1],argsort(ws1),1)
                  if tx==None: tx = sort(list(Set(ws1)))
                  ws2 = concatenate([ws2,ws1])
                  dws2 = concatenate([dws2,dws1])
                  qs2 = concatenate([qs2,qs1])
                  ds2 = concatenate([ds2,ds1])
                  fs2 = concatenate([fs2,fs1])
                  bs2 = concatenate([bs2,bs1])
                  vs2 = concatenate([vs2,vs1])
                  ns2 = concatenate([ns2,ns1])
                  ps2 = concatenate([ps2,ps1])
               ws2,dws2,ds2,bs2,vs2,ns2,fs2,ps2,qs2 = take([ws2,dws2,ds2,bs2,vs2,ns2,fs2,ps2,qs2],argsort(ws2),1)
               ws = concatenate([ws,ws2])
               dws = concatenate([dws,dws2])
               qs = concatenate([qs,qs2])
               ds = concatenate([ds,ds2])
               bs = concatenate([bs,bs2])
               vs = concatenate([vs,vs2])
               ns = concatenate([ns,ns2])
               ps = concatenate([ps,ps2])
               fs = concatenate([fs,fs2])

         if len(ws):
            ws,dws,ds,bs,vs,ns,fs,ps,qs = take([ws,dws,ds,bs,vs,ns,fs,ps,qs],argsort(ws),1)
            ws0 = 1*ws
            ws,dws,ds,bs,vs,ns,fs,ps,qs = adjspac([ws,dws,ds,bs,vs,ns,fs,ps,qs])
            xb,xe = extrema(ws)
            print xb,xe
            use = between(ws,w1,w2,inc=1)*greater(ps,1e-4)
            dfit = fs
#            dfit = divz(VTKGauss(ds*ns**2,1*(A*sumnorms/2),numret=1),\
#                      VTKGauss(ns**2,1*(A*sumnorms/2),numret=1))

            if scut < 100: wit = 4
            else: wit = 1
            wns = ns
            kx = 1
            for jm in range(wit):
               dresids = divz(ds-dfit,sqrt(vs))
               dresid = compress(between(ws,w1,w2),dresids)
               rchi2 = sum(dresid**2)/(len(dresid)-len(tx))
               print "Reduced chi^2 =",rchi2
               if jm and (scut < 100 and haveqe): wms = Lp(dresids,w=1)
               else: wms = ones(dresids.shape,float32)
               if haveqe: wns = ns * wms
               else: wns = ns * wms
               wns = where(wns,wns,min(compress(wns,wns))*1e-10)
               tx = wbins
               ntx = diff(searchsorted(ws0,tx))
               tx = wbins[:-1]+dbins[:-1]/2.0
#               tx = wbins[:-1]
               utx = 1.0*greater(ntx,median(compress(ntx,ntx))/2.)
               tx = compress(utx,tx)
               tx = compress(between(tx,xb,xe,inc=1),tx)
               if len(tx) < len(wbins)/2.:
                   print "Giving up! Not enough data in this one!",
                   print len(tx),len(wbins)
                   print ntx
                   print utx
                   return
               stck = splrep(ws,ds,w=wns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
               dfit = splev(ws0,stck)

            dresids = divz(ds-dfit,sqrt(vs))
            dresid = compress(between(ws,w1,w2),dresids)
            rchi2 = sum(dresid**2)/(len(dresid)-len(tx))
            print "Reduced chi^2 =",rchi2

            aresid = sqrt(divz(VTKGauss(wns*dresids**2,2*(A*sumnorms),numret=1), VTKGauss(wns,2.*(A*sumnorms),numret=1)))

            etck = splrep(ws,vs,w=ns,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
            btck = splrep(ws,bs,w=ns/ps,s=None,task=-1,k=kx,t=tx,xb=xb,xe=xe)
            ptck = (concatenate([ntx[:1],ntx]) + concatenate([ntx,ntx[-1:]]))/2.0

            sbinss = A*splev(wbins,stck)*between(wbins,xb,xe)
            bbinss = A*splev(wbins,btck)*between(wbins,xb,xe)
            efit = sumnorms*A*splev(wbins,etck)*between(wbins,xb,xe)*mean(ps)

            ebinss = getsqrt(efit)

            gtx = greater(VTKImageShift(concatenate([utx,[utx[-1]]]),-0.5,0.0,numret=1,interp=1),0.25)

            sbins2 = sbinss*gtx
            ebins2 = ebinss*gtx
            bbins2 = bbinss*gtx

       else:
          stck = splrep(wbins,sbins2/dbins,s=0,k=3)

       for r in ro:
#        for mpass in range(1+2*(pit < npit-1)):

        if momfix:
           if not data[r].has_key("pmh") or not len(data[r]["pmh"]):
              print name
              raise "Hermite Moments for %s not in %s" % (name,r)
           if haveqe:
              YHOrder = len(data[r]["pmh"])
              WHOrder = 0
           else:
              YHOrder = int(h+1)
              WHOrder = 4
        else:
           if haveqe:
              YHOrder = int(h/2+1)
              WHOrder = 0
           else:
              YHOrder = int(h+1)
              WHOrder = 0

        for mpass in range(2): 
          if mnorm and norms[r]:
             d[r]["fs0"] = d[r]["fs"]
             ps = d[r]["ps"]
             rfs = splev(d[r]["ws"],stck)
             rfs = rfs * data[r]["t"] * d[r]["dws"] * d[r]["qs"]
             if norms[r]: rfs = rfs/norms[r]
             d[r]["fs"] = rfs * ps
             for jab,ab in enumerate("ab"[:unsh+1]):
                ws = data[r][ab]["ws"].ravel()
                ss = zeros(ws.shape,float64)
                order = argsort(ws.ravel())
                put(ss,order,splev(take(ws,order),stck))
                ss = ss * between(ws,w1,w2,inc=1)
                ss = reshape(ss,data[r][ab]["ds"].shape)
                ss = ss * data[r][ab]["dws"] * data[r][ab]["qs"] * data[r]["t"]
                if norms[r]: ss = ss / norms[r]
                data[r][ab]["rfs"] = ss
                data[r][ab]["fs0"] = data[r][ab]["fs"]
                data[r][ab]["fs"] = data[r][ab]["rfs"] * data[r][ab]["ps"]
             if pit < npit-1 and not momfix:
                print "Let's try a new set of moments for",r
                ws = concatenate([data[r][ab]["ws"].ravel() for ab in "ab"[:unsh+1]])
                dws = concatenate([data[r][ab]["dws"].ravel() for ab in "ab"[:unsh+1]])
                xs = concatenate([data[r][ab]["xs"].ravel() for ab in "ab"[:unsh+1]])
                ys = concatenate([data[r][ab]["ys"].ravel() for ab in "ab"[:unsh+1]])
                ds = concatenate([data[r][ab]["ds"].ravel() for ab in "ab"[:unsh+1]])
                bs = concatenate([data[r][ab]["bs"].ravel() for ab in "ab"[:unsh+1]])
                ns = concatenate([data[r][ab]["ns"].ravel() for ab in "ab"[:unsh+1]])
                ls = concatenate([data[r][ab]["ls"].ravel() for ab in "ab"[:unsh+1]])
                rs = concatenate([data[r][ab]["rfs"].ravel() for ab in "ab"[:unsh+1]])
                fs = concatenate([data[r][ab]["fs"].ravel() for ab in "ab"[:unsh+1]])
                ps = concatenate([data[r][ab]["ps"].ravel() for ab in "ab"[:unsh+1]])
                qs = concatenate([data[r][ab]["qs"].ravel() for ab in "ab"[:unsh+1]])
                keep = 1.0*ls*greater(ds,-3*ns)*greater(ns,0)*not_equal(ds,0)*not_equal(qs,0)*not_equal(rs,0)
                if nsh: keep = keep*less(abs(ys), data[r]["nsh"])
                ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs = [compress(keep,a) for a in (ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs)]
                order = argsort(ws)
                ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs = [take(a,order) for a in (ws,dws,xs,ys,ds,bs,ns,fs,ps,rs,qs)]
                es = divz((ds-fs)/ns)

#                wxbasis = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ

#                m1b = polynom(ws/1e4,len(data[r]["pm1"])-1)
#                m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,len(data[r]["pm1"])-1)
                m1b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(data[r]["pm1"])-1)
                m1y1 = dot(data[r]["pm1"],m1b)

                wt = getsqrt(ds)*less((es/scut)**2,1.0**2)*between(ws,w1,w2,inc=1)
                wt = wt*less(abs(ys),h+1)
#                wt = wt*less(abs(ys-m1y1),h+1)
#                wt = wt * exp(-0.5*(ys-m1y1)**2/(h/2.)**2)

                if scut > 100: W1Order = max([1,len(data[r]["pm1"]) - 1])
                else: W1Order = max([[0,1][pit>0],len(data[r]["pm1"]) - 1])
#                m1b = polynom(ws/1e4,W1Order)
#                m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W1Order)
                m1b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,W1Order)
                m1 = svdfit(transpose(m1b)*wt[::,NewAxis],ys*wt)
                print m1
                data[r]["pm1"] = m1
                dy1 = dot(data[r]["pm1"],m1b)
                yc = ys - dy1

#                wt = getsqrt(ds)*less(es**2,scut**2)*less(abs(yc),h+1)*less(abs(yc),2.0*data[r]["pm2"][0]])
                wt = getsqrt(ds)*less((es/scut)**2,1.0)*between(ws,w1,w2,inc=1)
                if scut < 100 or (haveqe and momfix):
                   swt = wt*less(abs(yc),2.5*data[r]["pm2"][0])*less(abs(yc),h+1)
                else:
                   swt = wt
#                swt = wt
#                swt = wt*less(abs(yc),min([h+1,2.5*data[r]["pm2"][0]]))
                if scut > 100: W2Order = max([1,len(data[r]["pm2"]) - 1])
                else: W2Order = max([0,len(data[r]["pm2"]) - 1])
#               m2b = polynom(ws/1e4,W2Order)
#                m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
                m2b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,W2Order)
                is2p = 10
#                if haveqe and scut < 100: is2p = 1
                for s2p in range(10):
                   m2 = svdfit(transpose(m2b)*swt[::,NewAxis],power(yc,2)*swt)
                   sm2 = clip(getsqrt(dot(m2,m2b)),1,h)
                   swt = wt*less(abs(yc),3.0*sm2)
                data[r]["pm2"] = m2
                dy2 = clip(getsqrt(dot(m2,m2b)),1,h)
                print m2,quartiles(dy2)
                yn = divz(yc,dy2)
                gy = exp(-0.5*power(yn,2))
                mh = None
#                WHOrder,YHOrder,mh = 0,0,array([1.0])

 #                try:
                if norms[r]:
                   wt = divz(1.0,ns)*between(ws,w1,w2,inc=1)*not_equal(ds,0.0)*less((es/scut)**2,1.0)
                   if scut < 100 or (haveqe and momfix):
                      wt = wt*less(abs(yn),2.5)
                   print h,YHOrder
                   mhyb = hermite(yn,YHOrder,norm=1)
                   uds = ds
                   if scut < 100 or (haveqe and momfix):
                      mhyb = mhyb[1:]
                      uds =  (ds-rs*gy)
#                   mhwb = polynom(ws/1e4,WHOrder)
#                   mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
                   mhwb = legendre((ws-LDPCOEFCi)/LDPCOEFSi,WHOrder)
                   rs = splev(ws,stck)*dws*qs*data[r]["t"]
                   if norms[r]: rs = rs/norms[r]
                   mhb = [(rs*gy*wt*hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
                   mh = svdfit(transpose(mhb),wt*uds)
                   print mh
#                   mhb = [(hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
#                   gh = dot(mh,mhb)*gy
#                   if scut < 100 or (haveqe and momfix): gh = gh + gy
                else:
                   if scut < 100 or (haveqe and momfix):
                      mh = zeros(((YHOrder)*(WHOrder+1),),float64)
                   else:
                      mh = zeros(((YHOrder+1)*(WHOrder+1),),float64)
                      mh[0] = sqrt(pi)
#                except:
#                   mh = None
                data[r]["pmh"] = mh

                for ab in "ab"[:unsh+1]:
                   ws = data[r][ab]["ws"].ravel()
                   ys = data[r][ab]["ys"].ravel()
#                   wxbasis = legendre((ws-LDPCOEFCi)/LDPCOEFSi,len(LDPCOEF)-1)
#                   dcx0 = LDPZ*log(dot(LDPCOEF,wxbasis))+LDPZ
#                   m1b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W1Order)
                   m1b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,W1Order)
#                   m1b = polynom(ws/1e4,W1Order)
                   dy1 = dot(data[r]["pm1"],m1b)
                   yc = ys - dy1
                   data[r][ab]["yc"] = reshape(yc, data[r][ab]["ys"].shape)
                   data[r][ab]["ys"] = data[r][ab]["yc"]
#                   m2b = polynom(ws/1e4,W2Order)
#                   m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
                   m2b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,W2Order)
                   dy2 = clip(getsqrt(dot(data[r]["pm2"],m2b)),1,h)
                   yn = divz(yc,dy2)
                   gy = exp(-0.5*power(yn,2))
                   ps = gy
                   if data[r]["pmh"]:
                      mhyb = hermite(yn,YHOrder,norm=1)
                      if scut < 100 or (haveqe and momfix): mhyb = mhyb[1:]
#                      mhwb = polynom(ws/1e4,WHOrder)
#                      mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
                      mhwb = legendre((ws-LDPCOEFCi)/LDPCOEFSi,WHOrder)
                      mhb = [(hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb]
                      ps = dot(data[r]["pmh"],mhb)*gy
                      if scut < 100 or (haveqe and momfix): ps = ps + gy
                   ps = reshape(ps, data[r][ab]["ps"].shape)
                   yc = reshape(yc, data[r][ab]["ps"].shape)
#                   qdump("psold.fits",data[r][ab]["ps"])
#                   qdump("psnew.fits",ps)
#                   ps = ps * less(abs(data[r][ab]["yc"]),h)
                   data[r][ab]["ps"] = ps

                   ws = data[r][ab]["ws"][0]
                   wdir = median(diff(ws)) < 0
                   if wdir: ws = ws[::-1]
                   rfs = splev(ws,stck)
                   if wdir: rfs,ws = rfs[::-1],ws[::-1]
                   rfs = rfs * data[r][ab]["dws"][0]*data[r][ab]["qs"][0]*data[r]["t"]/norms[r]
                   fs = rfs[NewAxis,::] * data[r][ab]["ps"]
                   use = between(data[r][ab]["ws"],w1,w2,inc=1)
                   data[r][ab]["fs"] = fs*use
#                   qdump("data.fits",use*data[r][ab]["ps"])
                   fuse = use*less_equal(abs(yc),h)*data[r][ab]["ls"]*data[r][ab]["ns"]
                   x = divz(data[r][ab]["ds"] - data[r][ab]["fs"], data[r][ab]["ns"]).ravel()
                   x = compress(fuse.ravel(),x)
                   x2 = sum(power(x,2))
                   ndeg = len(x.ravel())-len(tx)-len(data[r]["pmh"])-len(data[r]["pm1"])-len(data[r]["pm2"])
                   print x2,len(x.ravel()),ndeg,divz(x2,ndeg)
#                   fs =  data[r][ab]["fs"]
#                   qdump("data.fits",data[r][ab]["ds"])
#                   qdump("modelold.fits",data[r][ab]["fs0"])
#                   qdump("model.fits",data[r][ab]["fs"])
##                   qdump("model.fits",fs)
#                   qdump("noise.fits",data[r][ab]["ns"]*reshape(less(abs(yc),h+1),data[r][ab]["ns"].shape))
#                   qdump("wave.fits",data[r][ab]["ws"])
#                   data[r][ab]["fs"] = fs
#                   data[r][ab]["fs"] = data[r][ab]["rfs"] * data[r][ab]["ps"]

          elif not momfix:
             data[r]["pmh"] = None

       final_pm2 = None
       final_pmh = None
       if nro:
          model = {}
          dymod = 0.1
          ws = wbins
          ys = arange(-2*h,2*h+dymod/2.,dymod,float32)
          ws,ys = ws[NewAxis,::]*ones(ys.shape)[::,NewAxis],ys[::,NewAxis]*ones(ws.shape)[NewAxis,::]
          ws = ws.ravel()
          yc = ys.ravel()
          W2Order = len(data[r]["pm2"]) - 1
#          m2b = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,W2Order)
          m2b = legendre((ws-LDPCOEFCi)/LDPCOEFSi,W2Order)
          pm2 = sum([norms[r]*data[r]["pm2"] for r in ro if norms[r]])
          if sumnorms: pm2 = pm2/sumnorms
          dy2 = clip(getsqrt(dot(data[r]["pm2"],m2b)),1,h)
          yn = divz(yc,dy2)
          gy = exp(-0.5*yn**2)
          ps = gy
#          print [(r,data[r]["pmh"]) for r in ro if norms[r] if data[r]["pmh"]]
          pmh = sum([norms[r]*data[r]["pmh"] for r in ro if norms[r] if sum(data[r]["pmh"])])
          if sumnorms: pmh = pmh/sumnorms
#          print pmh
          if pm2: final_pm2 = pm2
          if sum(pmh): final_pmh = pmh
          if sum(pmh):
             mhyb = hermite(yn,YHOrder,norm=1)
             if scut < 100 or (haveqe and momfix): mhyb = mhyb[1:]
#             mhwb = legendre((1e4/ws-LDPCOEFC)/LDPCOEFS,WHOrder)
             mhwb = legendre((ws-LDPCOEFCi)/LDPCOEFSi,WHOrder)
             mhb = asarray([(gy*hyb1*hwb1) for hyb1 in mhyb for hwb1 in mhwb])
             ps = dot(pmh,mhb)
             if scut < 100 or (haveqe and momfix): ps = ps + gy
          ps = reshape(ps, ys.shape)
          rfs = splev(wbins,stck)
          nfs = splev(wbins,etck)
          bfs = splev(wbins,btck)
          rfs = rfs[NewAxis,::] * ps
          efs = nfs[NewAxis,::] * ps * ps
          bfs = bfs[NewAxis,::] * ones(ps.shape,float64)
          sbins2 = sum(rfs*dymod)
          ebins2 = getsqrt(divz(sum(efs*dymod),divz(ptck,sum(ps*dymod))))
          bbins2 = divz(sum(bfs*dymod))

       if inter or (pit == npit-1):
          if inter: pdev = PGPLOT_DEV
          else: pdev = "%s/%s_1D.png/png" % (output,name)
          opendev(pdev,nx=3,ny=1,size=12,land=0,asp=1./3.)

          if scut > 100 and not haveqe:
               newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],getmar((dbins*sbins2+dbins*ebins2),50,150),ch=2)
               xlabel("\gl (\A)",ch=2)
               ylabel("PV weighted Counts/s",ch=2)
               subtitle("%s (p=%d)" % (name,1+pit),ch=2)
               rci = 1
               for r in ro:
                  if norms[r]:
                     rci += 1
                     wp,dwp,sp,dsp = compress(weights[r],[wbins,dbins,dbins*individ[r][0],dbins*individ[r][0]],1)
                     drawpt(wp,norms[r]*sp,ps=1,ch=2,ci=rci)
                     drawpterr(wp,dwp/2.,norms[r]*sp,norms[r]*dsp,ch=2,ci=rci)
               drawpt(wbins,dbins*sbins2,ps=1,ch=2,ci=1)
               drawpterr(wbins,dbins/2.,dbins*sbins2,dbins*ebins2,ch=2,ci=1)

          else:
               newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],multiply(abs(array(getmar((bbins2+2*ebins2)/2.,50,400))),[-1,1]),ch=2)
               xlabel("\gl (\A)",ch=2)
               ylabel("PV weighted Counts/A/s",ch=2)
               rci = 1
               drawpt(wbins,bbins2,ps=1,ch=2,ci=1)
               drawpterr(wbins,dbins/2.,bbins2,ebins2,ch=2,ci=1)
               drawln(wbins,ebins2,ls=1,ci=1)

          newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],multiply(abs(array(getmar((sbins2+2*ebins2)/2.,50,400))),[-1,1]),ch=2)
          xlabel("\gl (\A)",ch=2)
          ylabel("PV weighted Counts/A/s",ch=2)
          rci = 1
          for r in ro:
             if norms[r]:
                rci += 1
                wp,dwp,sp,dsp = compress(weights[r],[wbins,dbins,individ[r][4],individ[r][5]],1)
                drawpt(wp,norms[r]*sp,ps=1,ch=2,ci=rci)
                drawpterr(wp,dwp/2.,norms[r]*sp,norms[r]*dsp,ch=2,ci=rci)
#          drawpt(wbins,sbins2,ps=1,ch=2,ci=1)
#          drawpterr(wbins,dbins/2.,sbins2,ebins2,ch=2,ci=1)
#          drawln(wbins,ebins2,ls=1,ci=1)

          newbox([LDPDISP["lambda"][0],LDPDISP["lambda"][-1]],multiply(abs(array(getmar((sbins2+2*ebins2)/2.,50,400))),[-1,1]),ch=2)
          xlabel("\gl (\A)",ch=2)
          ylabel("PV weighted Counts/A/s",ch=2)
          drawpt(wbins,sbins2,ps=1,ch=2,ci=1)
          drawpterr(wbins,dbins/2.,sbins2,ebins2,ch=2,ci=1)
          drawln(wbins,ebins2,ls=1,ci=1)

          curs = ["O"]
          while inter and curs[0] != "g":
             if curs[0] in "qQI": os._exit(0)
             else: curs = getcurs(0,0)
          closedev()

    results = {}
    results["w"] = wbins
    results["dw"] = dbins
    results["s"] = sbins2*tt
    results["ds"] = ebins2*tt
    results["b"] = bbins2*tt
    results["tt"] = tt
    results["A"] = A

#    if not haveqe:
#       sbins2 = divz(sbins2*dbins)
#       ebins2 = divz(ebins2*dbins)
#       bbins2 = divz(bbins2*dbins)

    D = asarray([wbins,dbins,sbins2,ebins2,bbins2])

    naxes = array(D.shape)[::-1].tolist()
    F = FITS("%s/%s.fits" % (output,name),create=1)
    F.newhead("T",-32,len(naxes),naxes)
    F.modkey("object", name)
    F.modkey("exptime", tt)
    F.modkey("aperture", A)
    F.modkey("mnorm", mnorm)
    [F.modkey("norm%d" % (j+1), "(%s,%f)" % (r,norms[r])) for j,r in enumerate(ro)]
    if final_pm2:
        F.modkey("W2O", W2Order)
        [F.modkey("pm2%02d" % (j), final_pm2[j]) for j in range(len(final_pm2))]
    if sum(final_pmh):
        F.modkey("WHO", WHOrder)
        F.modkey("YHO", YHOrder)
        [F.modkey("pmh%02d" % (j), final_pmh[j]) for j in range(len(final_pmh))]
    F.putdata(D)
    F.close()

    return results

def PullOneGalaxy(object,robjects,objno,nbin=1,output="Spectra",inter=False,scut=5,w1=0,w2=0,hap=0,momfix=0):
   data = ReadOneObject(object,robjects,objno)
   t0 = time.time()
   FitOneObject(data,nbin=nbin,output=output,name=object,inter=inter,scut=scut,w1=w1,w2=w2,hap=hap,momfix=momfix)
   t1 = time.time()
   print "Elapsed time was %.3fs" % (t1-t0)

