#!/usr/bin/env python

from math import atan2
from sys import stdout
from os import remove
import time
import tempfile
from ephem import *
from masks import *
from numpy.oldnumeric.mlab import *
from basics import *
from numpy.core.numerictypes import *
import numpy
from VTKHelperFunctions import *
from FITS import *
from NEWFIRM import JTRANS
import SimpleQueue
from sets import Set

import ZODB, ZODB.FileStorage, transaction
import persistent
from persistent.dict import *
from persistent.list import *
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree

NSCALE = 0.4

class newfirm(persistent.Persistent):
    def __init__(self,record,keys):
       if type(record) != ListType:
           simplify = lambda l: l.replace(" ",",")
           record = simplify(record)
           simplify = lambda l: l.replace(",,",",")
           while record.count(",,"):
              record = simplify(record)
           record = record.replace(",',","',")
           record = record[:-3]+"'"
           record = record.replace("'0,[s]'","0.0")
           record = record.replace(",[s]'","'")
           record = record.split(",")
           record[1:] = map(eval, record[1:])
       self.info = {}
       [self.info.setdefault(key,rec) for rec,key in zip(record,keys)]
       if self.info["NOCCOADD"] == "N/A":
          self.info["NOCCOADD"] = 1
       if type(self.info["FILTER"]) not in [StringType]:
          self.info["FILTER"] = str( self.info["FILTER"])
       if type(self.info["RAOFF"]) in [StringType] and self.info["RAOFF"] != "N/A":
          self.info["RAOFF"] = float(self.info["RAOFF"].split()[0])
       if type(self.info["DECOFF"]) in [StringType] and self.info["DECOFF"] != "N/A":
          self.info["DECOFF"] = float(self.info["DECOFF"].split()[0])
       self.base = os.path.basename(self.info["Filename"]).replace(".fits","")
       self.name = os.path.basename(self.info["Filename"])
       self.bdir = os.path.dirname(self.info["Filename"])
#       print self.name,
#       print self.info["RA"],
#       print self.info["DEC"]
       if self.info["RA"] != "N/A" and self.info["DEC"] != "N/A" and self.info["MJD-OBS"] != "N/A":
          self.ra = hrrad(scanSexagesimal(self.info["RA"]))
          self.dec = degrad(scanSexagesimal(self.info["DEC"]))
          rao = self.info["RAOFF"]/cos(self.dec)
          deco = self.info["DECOFF"]
          self.mjd = self.info["MJD-OBS"]
          self.when = int(self.mjd)
       else:
          self.ra = 0.0
          self.dec = 0.0
          rao = 0.0
          deco = 0.0
          self.mjd = 0.0
          self.when = 0
       rint = 10.0*60.0
       ira0 = int((raddeg(self.ra)*3600-rao)/rint+0.5)
       idec0 = int((raddeg(self.dec)*3600-deco)/rint+0.5)
       self.ra0 = ira0*rint/3600.0/15.0
       self.dec0 = idec0*rint/3600.0
       self.RA0 = formatSexagesimal(self.ra0,4,3600).strip()
       self.DEC0 = formatSexagesimal(self.dec0,4,3600).strip()
       if self.DEC0[0] != "-": self.DEC0 = "+"+ self.DEC0
       if abs(raddeg(self.dec)) < 10: self.DEC0=self.DEC0[0] + "0"+self.DEC0[1:]
       if self.RA0.find(":") == 1: self.RA0 = "0"+self.RA0
       self.RA0 = self.RA0.replace(":","")
       self.DEC0 = self.DEC0.replace(":","")
       self.type = self.info["OBSTYPE"]
#       print self.name,
#       print self.info["FILTER"]
       if "Dark" in self.info["FILTER"] or "dark" in self.info["OBJECT"].lower(): self.type="dark"
       if "FOCUS" in self.info["NOCDHS"]: self.type="focus"
       if "flat" in self.info["OBJECT"].lower(): self.type = "flat"
       if self.type =="flat": self.pref = "flt"
       elif self.type =="object": self.pref = "obj"
       elif self.type =="focus": self.pref = "foc"
       elif self.type =="dark": self.pref = "drk"
       else: self.pref = "unk"
       if self.pref == "obj": self.object = "%s%s%s_%s" % (self.pref,self.RA0,self.DEC0,self.info["FILTER"])
       elif self.pref == "flt": self.object = "flat"+self.info["FILTER"]+self.info["LAMPSTAT"]
#       elif self.pref == "flt": self.object = self.info["OBJECT"].lower()+self.info["LAMPSTAT"]
       elif self.pref == "foc": self.object = "focus"
       elif self.pref == "drk":
          self.object = "dark%sx%s" % (self.info["NOCTIM"],self.info["NOCCOADD"])
       elif self.pref == "unk": self.object = "unknown"
       self.filter = self.info["FILTER"]
       self.jno = int(self.base[-4:])
       self.exptime = self.info["EXPTIME"]
       if self.exptime == "N/A":
          self.exptime = 0
       self.asecs = self.info["RAOFF"]
       self.dsecs = self.info["DECOFF"]
       self.newfile = self.base
       self.skys=[]
       self.groupt=[]
       self.groupm=[]
       self.groups=[]
       self.groupe=[]
       self.groupx=[]
       self.groupy=[]
       self.refgroups=[]
       self.sets=0
       mode1s = glob("%s/mode1/%sc?.fits.bad" % (self.object, self.name.replace(".fits","")))
       if mode1s: self.bad = True
       else: self.bad = False

    def __repr__(self):
      return "%s[%d(%d)]" % (self.newfile,self.exptime,self.info["NOCCOADD"])

    def order(self,no):
       self.norder = no

    def group(self,t,refdir=None):
       # obsoleted! use FindGroups!
       dx,dy= normalcoord(degrad(t.ra) % (2*pi),degrad(t.dec),\
               degrad(self.ra) % (2*pi),degrad(self.dec), conv=180.0*3600.0/pi)
       m0 = abs(dx) < 300
       m1 = abs(dy) < 300
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)
       if m0 and m1 and m2 and m3:
           if refdir: self.refgroups.append( refdir+"/"+t.newfile)
           else:
              self.groups.append( t.newfile)
              self.groupe.append( t.exptime)
              self.groupx.append( -dx/NSCALE)
              self.groupy.append( dy/NSCALE)

    def clean(self,t):
       if t != self and t.groups == self.groups: t.groups=[]

    def set(self,t):
       m0 = less(abs(self.mjd - t.mjd),0.4)
       m1 = (self.type == t.type)
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)
       if m0 and m1 and m2 and m3: self.sets += 1

def readdb(db,targ,kind):
   f=open(db,"r")
   df = f.readlines()
   f.close()
   header = df[1][1:].replace(" ","").replace("''","','")
   keys = eval(header)
   j = nonzero([(l[0] not in ["#","["]) for l in df])
   df = [df[k] for k in j]
   t = [newfirm(df[j],keys) for j in range(len(df))]
   t = [i for i in t if ((targ and targ in i.object) or not targ) and (kind and kind==i.pref or not kind)]
   if targ: [setattr(i,"object",targ) for i in t]
   return t

def reduce2groups(t):
   [[i.clean(x) for x in t] for i in t]
   u = nonzero([i.groups != [] for i in t])
   t = [t[j] for j in u]
   return t

DBKEYS = ["MJD-OBS","EXPTIME","RA","DEC","FILTER","RAOFF","DECOFF","OBSTYPE","OBJECT",
        "LAMPSTAT","NOCDHS","NOCTIM","NOCCOADD"]

def FindGroups(t,useint=0):
   n = len(t)
   ra = array([i.ra for i in t])
   dec = array([i.dec for i in t])
   fno = array([int(i.jno) for i in t])
   x, y = ra, dec
   dx = (x[::,NewAxis] - x[NewAxis,::])*cos(y)*180/pi*60.0
   dy = (y[::,NewAxis] - y[NewAxis,::])*180/pi*60.0
   m0 = less(abs(dx),10.0)
   m1 = less(abs(dy),10.0)
   m2 = array([[(t[i].filter==t[j].filter) for j in range(n)] for i in range(n)])
   m3 = array([[(t[i].type==t[j].type) for j in range(n)] for i in range(n)])
   U = m0*m1*m2*m3
   if useint:
      m4 = array([[(t[i].interval==t[j].interval) for j in range(n)] for i in range(n)])
      U = m4 * U
   t[0].groupt = [t[0]]
   t[0].groupm = [t[0].mjd]
   t[0].groups = [t[0].newfile]
   t[0].groupe = [t[0].exptime]
   t[0].groupx = [0.0]
   t[0].groupy = [0.0]
   G = [t[0]]
   I = [0]
   for j in range(n):
      new = 1
      if j in I:
         new = 0
      else:
         for i,g in zip(I,G):
            if t[j].newfile in g.groups:
              new = 0
            else:
              if U[j][i]:
                  new = 0
                  g.groupt += [t[j]]
                  g.groupm += [t[j].mjd]
                  g.groups += [t[j].newfile]
                  g.groupe += [t[j].exptime]
                  g.groupx += [-dx[i][j]]
                  g.groupy += [dy[i][j]]
      if new:
         left = 1
         t[j].groupt = [t[j]]
         t[j].groupm = [t[j].mjd]
         t[j].groups = [t[j].newfile]
         t[j].groupe = [t[j].exptime]
         t[j].groupx = [0.0]
         t[j].groupy = [0.0]
         G.append(t[j])
         I.append(j)
   return G

def reduce2groups(t):
   [[i.clean(x) for x in t] for i in t]
   u = nonzero([i.groups != [] for i in t])
   t = [t[j] for j in u]
   return t


def CollateDB(db,targ="",filt="",nume=0,kind=None,watching=False):
  if type(db) is StringType:
    print "Reading database",db
    ti = readdb(db,targ,kind)
  else:
    ti = [db[i] for i in db]

  t = copy.copy(ti)
  if len(t):
    mjd0 = int([i.mjd for i in t if i.mjd][0])
    mjd1 = 24*(t[-1].mjd-mjd0)
  else:
    return {}
  t = FindGroups(t)

  intervals = []
  for k in t:
    if k.pref == "obj" and len(k.groups)>1:
       dt = diff(k.groupm)*24*60
       bdt,sdt = mean(dt),rms(dt)
       mdt = max([bdt+3*sdt,3*median(dt)])
       g = nonzero(greater(dt,mdt))
       if len(dt) not in g: g = concatenate([g,[len(k.groupm)]])
       ng = len(g)
       for j in range(ng):
          if j: mjd = k.groupm[g[j-1]:g[j]]
          else: mjd = k.groupm[:g[j]+1]
          mjd = 24*(asarray(mjd) - mjd0)
          intervals.append(extrema(mjd))

  if intervals:
     dt = transpose(intervals)
     ct = 0.5*(dt[0]+dt[1])
     ot = argsort(ct)
     intervals = [intervals[j] for j in ot]
     dt = transpose(intervals)
     ct = 0.5*(dt[0]+dt[1])

     idt0 = dt[0].astype(int32)
     idt1 = (60*(dt[0]-idt0)).astype(int32)
     jdt0 = dt[1].astype(int32)
     jdt1 = (60*(dt[1]-jdt0)).astype(int32)

  t = copy.copy(ti)
  bad = []
  for j,i in enumerate(t):
      dti = (24*(i.mjd-mjd0))
      if intervals: dtest = less_equal(dti,dt[1]) * greater_equal(dti,dt[0])
      if i.pref == "obj":
         if intervals and i.mjd and dtest.any():
            l = nonzero(dtest)[0]
            if l == len(intervals)-1 and (i.mjd-t[j-1].mjd)<5*i.exptime:
               i.interval = ""
            else:
               i.interval = "%02d%02d_%02d%02d" % (idt0[l],idt1[l],jdt0[l],jdt1[l])
            i.object = "%s_%d_%s" % (i.object,mjd0,i.interval)
         else:
            i.interval = ""
         if not i.interval: bad.append(i.object)
      else:
             i.interval = ""

  FindGroups(t,useint=1)

  targets = list(Set([i.object for i in t if i.object not in bad]))

  groups = {}
  for target in targets:
      u = nonzero([((targ and targ in i.object) or (target in i.object)) for i in t])
      tu = [t[j] for j in u]
      if filt: tu = [i for i in tu if i.filter == filt]
      if (targ and targ in target or not targ) and len(tu)>=nume:
          print "Target %40s [%5d]" % (target,len(tu))
      [[i.set(x) for x in tu] for i in tu]
      [tu[j].order(j) for j in range(len(tu))]
      if len(tu)>=nume and (targ and targ in target or not targ):
         if targ: groups[targ] = tu
         else: groups[target] = tu

  if type(db) is not StringType:
    transaction.abort()

  if targ and targ in groups:
     return groups[targ]
  elif groups or watching:
     return groups
  else:
     raise NameError, "Target %s not found" % (targ)

def UpdateZODB(zodb,newfs=0,verb=0):
    ddir = os.path.dirname(zodb)
    bad = ddir+"/BAD"
    badness = {}
    if os.path.exists(bad):
      d = {}
      b = open(bad,"r")
      badness = d.fromkeys(b.read().split("\n"))
      b.close()
    if newfs: [os.remove(o) for o in glob(zodb+"*")]
    storage = ZODB.FileStorage.FileStorage(zodb)
    zdb = ZODB.DB(storage)
    conn = zdb.open()
    dbroot = conn.root()
    if newfs or "Data" not in dbroot:
       tree = OOBTree()
       dbroot["Data"] = tree
       transaction.commit()
    tree = dbroot["Data"]
    exists = [tree[i].info["Filename"] for i in tree]
    for i in tree:
       if i in badness:
           del tree[i]

    fitslist = [f for f in glob(ddir+"/image????.fits") if f not in exists]
    fitslist += [f for f in glob(ddir+"/image?????.fits") if f not in exists]
    fitslist += [f for f in glob(ddir+"/image??????.fits") if f not in exists]
    if not fitslist:
       fitslist = [f for f in glob(ddir+"/raw?????.fits") if f not in exists]
    fitslist = [f for f in fitslist if f.split(".")[-2][-1] in digits]
    fitslist.sort()
    for f in fitslist:
       if not badness.has_key(os.path.basename(f)):
          F = FITS(f)
          fi0 = [F[k] for k in DBKEYS]
          fi = []
          for i in fi0:
            if type(i) == StringType: fi.append(i.strip())
            else: fi.append(i)
          fr = [f]+fi
          F.close()
          nf = newfirm(fr,["Filename"]+DBKEYS)
          fb = os.path.basename(f)
          tree[fb] = nf
          transaction.commit()
          if verb:
             print "Added %s[%s]" % (fb,nf.object)

    transaction.commit()
    conn.close()
    zdb.close()
    storage.close()
#    os.remove(zodb+".lock")
#    print glob(zodb+"*")

if __name__ == "__main__":
    argv,targ = rdarg(argv,"-o",None,None)
    argv,kind = rdarg(argv,"-t",None,None)
    argv,nume = rdarg(argv,"-l",int,0)
    argv,filt = rdarg(argv,"-f",None,None)
    db = argv[1]

    CollateDB(db,targ,filt,nume=nume,kind=kind)

